(* Delphi-Unit (Unicode)
   Copy, pack and unpack files, create and read zip archives
   optional with encryption

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   Acknowledgements:
     ZLibEx (Vers. 1.2.3) and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Sept. 2005
   Vers. 2 - Oct. 2005   : Zip64 extensions
   Vers. 3 - Dec. 2005   : AE-2 encryption added (see AesLib)
   Vers. 4 - Apr. 2006   : Verify option added
   Vers. 4.1 - Jul. 2006 : some minor changes and corrections
   Vers. 4.2 - Aug. 2006 : new error: path too long, TZip.Destroy fixed (path>255)
   Vers. 4.3 - Sep. 2006 : exception in verify routines fixed
   Vers. 4.4 - Apr. 2007 : Warnings on setting attributes after copy
                           New error messages
   Vers. 4.5 - May  2007 : Verify on zip archives
   Vers. 4.6 - Jul. 2007 : allow setting of encryption depth on Zip files
   Vers. 4.7 - Oct. 2007 : alternate copy using SHFileOperation
                           without compression and encryption
   Vers. 4.8 - Jan. 2008 : full timestamp copy
   
   Vers. 6  - Feb. 2009 : Unicode Version for Delphi 2009
                          ===============================
     Notes: 1. GZip: The most recent file format description (4.3) comes from 1996 (RFC1952).
                     There is no Unicode or UTF-8 support specified. The actual
                     linux version of GZip writes filenames as UTF-8, but does
                     not use a flag as marker.
                     To retain compatibilty this unit writes filenames in two
                     different ways:
                     a) Filename is ISO-8859 compliant: as RFC1952
                        Op.Syst. is DOS (0), filename is AnsiString
                     b) Filename is not ISO-8859:
                        Op.Syst. is NTFS (11), filename is UTF8String
                     Reading files assumes AnsiString if Op.Syst. DOS, UTF8String
                     in all other cases
            2. Zip:  see PkZip App-Note Version: 6.3.2 (Sept. 2007)
                     uses general purpose flag bit 11
            3. XL:   all full file- and directory paths are expanded with prefix \\?\
   
   Vers. 6.1   - Jul. 2009 : Read zip files, Copy file storage
   Vers. 6.2   - Jan. 2011 : several minor changes and enhancements
   Vers. 6.3   - Sep. 2011 : optional setting of filename in gz header
   Vers. 6.4   - Oct. 2011 : new thread (TVerifyThread) to compare files
                             (plain, gz packed and/or encrypted)
                             New extension for gz-encrypted files
   Vers. 6.5   - Nov. 2011 : compressed and uncompressed size as int64 in 
                             gzip extrafield
   Vers. 6.6   - Aug. 2012 : changes to use zlib routines delivered with Delph XE2
   Vers. 6.7   - Jul. 2013 : changes in reading gz header and trailer
                             optional copying of ACL, CopyOptions instead of booleans
   Vers. 6.8   - Mar. 2014 : errors in code tables OEM <> ANSI fixed
                             automatic detection of crypting length in zip files
   Vers. 7.0   - Jan. 2015 : use of single file zip archives (alt. to gz) added
         7.1   - Jan. 2016 : reorganization of units (XlFileUtils, FileConsts)
                             setting of zip general purpose flag fixed
   Vers. 7.2   - Oct. 2016 : data format for encryption changed for enc and gze:
                             header includes info for encryption depth (old format is 
                             still supported for decryption)
   Vers. 8.0   - June 2017 : support of splitted zip files
         8.1   - Aug. 2017 : new enc data header holds timestamp and attribute
         8.2   - Feb. 2018 : optional copying of alternate data streams

   last updated: Feb. 2018
   *)

unit ExtFileTools;

interface

uses Winapi.Windows, System.Classes, System.SysUtils, System.Contnrs,
  ZLibEx, AesLib, FileUtils, CBFunctions;

const
  GzExt = '.gz';      // default extension for GZip
  GzeExt = '.gze';    // extension for encrypted GZip
  GzExtMask = '.gz*';
  ZipExt = '.zip';
  CrExt = '.enc';      // default extension for not compressed encrypted files

  (* Format of encrypted gz-files:
      Gzip-Header : 10 bytes as default
                    new: Flag byte: bit 5 = encrypted
      Extra field : (ID=1) Optional if size > 4GB  - 20 bytes
      Extra field : (ID=$524A) Signature JR + specification of the encryption depth - 6 bytes
      Filename    : ISO-8859-1 (OS=0 - FAT) or UTF-8 (OS=11 - NTFS)
      Enc-Header  : 10, 14 or 18 bytes (depending on the encryption depth)
      Enc-Data    : Same number of bytes as compressed original file
      Enc-Trailer : 10 bytes  (see above)
      Gzip-Trailer: 8 bytes
                    CRC always = 0

    Format of not compressed encrypted file
      Signature  : 4 bytes: JREx
      Enc-Header : 10, 14 or 18 bytes (depending on the encryption depth):
                   Salt value (8, 12 or 16 bytes) + password verification value (2 bytes)
      Enc-Data   : Same number of bytes as original file
      Enc-Trailer: 10 bytes Authentication code
  *)

  defBufferSize = 512*1024;     // default buffer size for copy operations

  smDenyNone = FILE_SHARE_READ or FILE_SHARE_WRITE;
  wmSequential = FILE_FLAG_SEQUENTIAL_SCAN;
  wmNoCache = FILE_FLAG_WRITE_THROUGH;

  InvalidAttributes = $FFFF;

  // error codes
  errOK         = 0;
  errFileCreate = 1;
  errFileOpen   = 2;
  errFileClose  = 3;
  errFileRead   = 4;
  errFileWrite  = 5;
  errFileAttr   = 6;  // Error setting attr.
  errFileFull   = 7;
  errFileGZip   = 8;  // ill. GZip header
  errFileCheck  = 9;  // corrupt copied or packed file
  errFileECrypt = 10; // error encrypting file
  errFileDCrypt = 11; // error decrypting file
  errFileVerify = 12; // verification error
  errLongPath   = 13; // path too long (>260)
  errNotFound   = 14; // source file not found
  errFileTS     = 15; // Error setting timestamp after copy
  errStorage    = 16; // Error copying document summary
  errTimeout    = 17; // Timeout error on file transfer
  errStream     = 18; // undefined stream
  errAcl        = 19; // Error copying ACL
  errFileExists = 20; // File already exists
  errSzMismatch = 21; // verify error (size mismatch)
  errVerOpen    = 22; // verify error (opening file)
  errDirCreate  = 23; // could not create directory
  errFtpRead    = 24; // error reading via FTP
  errFtpWrite   = 25; // error writing via FTP
  errFtpConnect = 26; // connection error on FTP
  errFtpBroken  = 27; // connection reset by server
  errFTPDatConn = 28; // error opening data connection
  errFtpTimeout = 29; // Timeout error on file transfer
  errCompare    = 30; // verify error (Error comparing files)
  errZipCrSeg   = 31; // Error creating new zip segment
  errSignature  = 32; // ill. Zip signature
  errExtract    = 33; // Error extracting file
  errFormat     = 34; // Unsupported file format
  errTmpFile    = 35; // Temp. file could not be renamed
  errZipRdSeg   = 36; // Error reading next zip segment
  errAltStreams = 37; // Error copying alternate file streams

  errUserBreak  = $80; // process stopped by user
  errAllCodes   = $00FF;

  errError      = $0100;
  errWarning    = $0200;
  errError2     = $0300;
  errAllTypes   = $0F00;

  errCopy       = $1000;
  errGZip       = $2000;
  errGUnzip     = $3000;
  errZip        = $4000;
  errEncrypt    = $5000;
  errDecrypt    = $6000;
  errUnzip      = $7000;
  errAllSources = $F000;

  // ZLib parameters
  defWindowBits = -15;  // for GZip format
  defMemLevel = 8;

  // crypt signature
  CrSignatur = $4245524A;

  // gz signature
  GzSignatur = $8b1f;
  Z_DEFLATE = 8;
  Z_LZMA = 14;   // LZMA

  // gzip operating system
  OsDos  = 0;
  OsUnix = 3;
  OsNTFS = 11;

  // gzip flag byte
  gfASCII      = $01; { bit 0 set: file probably ascii text }
  gfHeaderCrc  = $02; { bit 1 set: header CRC present }
  gfExtraField = $04; { bit 2 set: extra field present }
  gfFileName   = $08; { bit 3 set: original file name present }
  gfComment    = $10; { bit 4 set: file comment present }
  gfEncrypted  = $20; { bit 5 set: AE-2 encryption used for data after compressing }
  gfReserved   = $C0; { bits 6..7: reserved }

  ExtraGz64IDA = 1;         // Extra field tag for 64 size info (alternative value)
  ExtraGz64ID  = $464C;     // Extra field tag for 64 size info ('LF')
  ExtraGzCrypt = $524A;     // Extra field tag for encryption key mode ('JR')

  // PkZip signatures
  PkLocalSignatur = $04034b50;
  PkDirSignatur = $02014b50;
  PkEndSignatur = $06054b50;
  Pk64RecSignatur = $06064b50;
  Pk64LocSignatur = $07064b50;
  PkSplitSignatur = $08074b50;

  // PkZip Constants
  VersDef = $14;          // Unicode requires WinZip Vers. 12
  VersZip64 = $2D;
  VersDos = 0;            // VersWin = $A0
  zipEncryptFlag = 1;     // File is encrypted
  zipMaxDeflFlag = 2;     // Deflating - Maximum (-exx/-ex) compression option was used.
  zipPbFlag      = $80;   // Unused flag - set for PB compression as single file
  zipUtf8Flag    = $800;  // Filename and comment are UTF-8 coded
  zipNoComp      = 0;     // File is stored
  zipDeflate     = 8;     // File is deflated
  zipLzma        = 14;    // Use LZMA compression
  zipEncrypted   = 99;    // Compression methode uses encryption
  ExtraZip64 = 1;         // Extra field tag for Zip64
  ExtraCrypt = $9901;     // Extra field tag for encryption
  VendorID   = $4541;     // 'AE' - see: http://www.winzip.de/aes_info.htm
  VendorVersAE1 = 1;
  VendorVersAE2 = 2;
  AE1Size       = 20;     // see: http://www.winzip.de/aes_info.htm#crc-faq
  AE2Mode128 = 1;
  AE2Mode192 = 2;
  AE2Mode256 = 3;

  tmpGzipName = 'gzip.tmp';
  tmpEncName  = 'encrypt.tmp';
  tmpUzipName = 'unzip.tmp';
  tmpCompName = 'compare.tmp';
  tmpZipName  = 'zip.tmp';
  UMaxMemSize = 50*1024*1024;  // max. size of temp. memory stream (uncompressed)
  CMaxMemSize = 1024*1024;  // max. size of temp. memory stream (compressed)
  MaxCardinal = $FFFFFFFF;    // max. value for cardinals (4 bytes)

type
  TCopyOption = (coCompress,coZipCompress,coVerify,coOverwrite,coStorage,coAltStreams,
    coPermissions,coAttributes,coTimeStamp,coCreationTime,coCheckTime,coNoCache);
  TCopyOptions = set of TCopyOption;

  TZipState = (zsSingle,zsMulti,zsError);


  TGzHeader = packed record
    Signature : word;
    Methode,
    Flags     : byte;
    TimeStamp : cardinal;
    XFlags,
    OpSys     : byte;
    end;

  TGzExtra64 = packed record
    SubfieldID,
    Size        : word;
    CSize,USize : int64;
    end;

  TGzExtraCrypt = packed record
    SubfieldID,
    Size,
    KeyMode     : word;
    end;

  TGzFileInfo = record
    Filename  : string;          // original filename
    UnixTime  : cardinal;        // original timestamp as Unix time
    Attr      : word;            // original file attributes
    Count,
    CSize,
    USize     : int64;           // original size
    Encrypted : boolean;         // file is encrypted
    EncMode   : integer;         // encryption depth
    end;

  TEncHeader = packed record
    Signature,
    TimeStamp : cardinal;        // original timestamp as Unix time
    Attr      : word;            // original file attributes
    end;

  TExtraFieldInfo = packed record
    Tag,ESize   : word;
    end;

  TPKExtraLocalField = packed record
    Tag,ESize : word;
    Buffer    : array [0..3] of int64;
    end;

  TPKExtraEncField = packed record
    Tag,ESize,
    Version,Vendor  : word;
    Mode            : byte;
    Method          : word;
    end;

  TPKLocalHeader = packed record
    Signatur    : cardinal;
    Vers,Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength : word;
    end;

  TPKDirHeader = packed record
    Signatur    : cardinal;
    VersMade,
    VersExtr,
    Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength,
    CommLength,
    DiskStart,
    IntAttr     : word;
    ExtAttr,
    Offset      : cardinal;
    end;

  TPKEndHeader = packed record
    Signatur    : cardinal;
    ThisDisk,
    StartDisk,                  // Start of central directory
    ThisEntries,
    TotalEntries : word;
    DirSize,
    Offset      : cardinal;
    CommLength  : word;
    end;

  TPKZip64CdRecord = packed record
    Signatur    : cardinal;
    Size        : int64;
    VersMade,
    VersExtr    : word;
    ThisDisk,
    StartDisk   : cardinal;
    ThisEntries,
    TotalEntries,
    DirSize,
    Offset      : int64;
    end;

  TPKZip64CdLocator = packed record
    Signatur    : cardinal;
    StartDisk   : cardinal;
    Offset      : int64;
    TotalDisk   : cardinal;
    end;

  TZipFileInfo = record   // data from local header
    Filename        : string;
    Flag,Method     : word;
    TimeStamp,Crc   : cardinal;
    CSize,USize     : int64;
    KeyMode         : integer;
    end;

  TPkFileRef = class(TObject)
    Offset  : int64;
    constructor Create (Ofs : int64);
    end;

  EZipError = class(EInOutError)
  public
    ZipError,SysError : integer;
    SysErrorMsg       : string;
    constructor Create (const ErrString : string; ErrCode : integer;
                        const ASysErrMsg : string = '');
    end;

  TPkHeader = class(TObject)
    FName           : RawByteString;   // Original filename (OEM or UTF8)
	  ErrCode,
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethod,
    Tag,DirLength,
    Segment,EncMode,
    EncVers,EncMeth : word;
    Offset,
    CSize,USize     : int64;
    IsDir,IsAE2,
    Zip64           : boolean;
    constructor Create (const Fn : RawByteString; Ec,Al,Ts,Chk : cardinal;
                        AA,AF,AM,EV,ED,EM,SM : word; Ofs,CSz,USz : int64);
    constructor Assign (APkHeader : TPkHeader; Ofs : int64);
    end;

const
  // Extra field for AES encryption (see: http://www.winzip.de/aes_info.htm)
  defPkExtraCrypt : TPKExtraEncField = (Tag : ExtraCrypt; ESize : 7; Version : VendorVersAE2;
                       Vendor : VendorID; Mode : AE2Mode128; Method : zipDeflate);

type
  TShellCopyMode = (scmCopy,scmQuiet);
  TShellCopyModes = set of TShellCopyMode;

  EWriteFileStreamError = class(Exception)
  private
    FCode : integer;
  public
    constructor Create (ECode : integer; const EMsg : string);
    property ErrorCode : integer read FCode;
    end;

  // Filestreams with sequential write
  TWriteFileStream = class(TFileStream)
    constructor Create (const AFileName: string; Overwrite : boolean = true;
                        WriteMode : cardinal = wmSequential; ShareMode: Word = 0);
    function Write(const Buffer; Count: Longint): Longint; override;
    end;

  // copy stream in own thread
  TCopyStreamThread = class (TThread)
  private
    FOnSProgress     : TFileProgressEvent;
    FError           : integer;
    FSysErrMsg       : string;
    FLength,
    FTotal           : int64;
    function GetDone : boolean;
    procedure SetProgress (ACallBack : TFileProgressEvent);
  protected
    FBufSize         : cardinal;
    FBuffer          : array of byte;
    sSource,sDest    : TStream;
    FUserBreak       : boolean;
    FAction          : TFileAction;
    FCount           : int64;
    procedure UpdateProgress;
    procedure DoProgress (AAction : TFileAction; ACount : int64);
    function CopyStream : integer;
    procedure EndThread (Error : integer);
  public
    constructor Create (ASource,ADest : TStream; ALength : int64;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure Execute; override;
    property Done  : boolean read GetDone;
    property ErrorCode : integer read FError;
    property OnProgress : TFileProgressEvent read FOnSProgress write SetProgress;
    property SysErrorMsg : string read FSysErrMsg;
    end;

  // base class for all copy functions (compressed and crypted)
  TCopyThread = class (TThread)
  private
    FOnProgress      : TFileProgressEvent;
    FCheckTime       : boolean;           // check for illegal filetime
    FShellCopy       : TShellCopyModes;   // use shell copy function, overrides compression and encryption
    FCompress        : boolean;
    FSysError        : cardinal;
    function CheckSysError (ASysError : cardinal) : boolean;
    function GetDone : boolean;
    procedure SetNoWriteCache (Value : boolean);
    procedure SetProgress (ACallBack : TFileProgressEvent);
    procedure SetTempDirectory (const ATempDir : string);
  protected
    FError,FErrFlag  : integer;
    FSysErrMsg,
    SourceName,
    DestName,
    SourceNameXL,
    DestNameXL,
    fTmpDir          : string;
    FBuffer          : array of byte;
    sSource,sDest    : TStream;
    FMaxWait,
    FWriteMode,
    FBufSize         : cardinal;
    FUserBreak,
    FCopyTimeStamp,
    FCopyAttr,
    FCopyCrTm,
    FCopyStorage,
    FCopyAltStreams,
    FCopyAcl,
    FOverwrite,
    FVerify          : boolean;
    FCompLevel       : TZCompressionLevel;
    FCompType        : TZStrategy;
    FCrc             : cardinal;
    UTotal,CTotal    : int64;
    FRatio           : integer;
    FCrypting        : TEncryption;
    FAction          : TFileAction;
    FCount           : int64;
    FSpeedLimit      : boolean;
    function CopyStream (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean = false) : integer;
    function CompareStreams (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer;
    function GzCompress (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer;
    function GzUnCompress (fSource,fDest : TStream; ASpeedLimit : boolean = false) : integer; overload;
    function GzUnCompress (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean = false) : integer; overload;
    function EncryptFile : integer;
    function ReadCryptHeader(fSource : TStream; var EncHd : TEncHeader) : integer; overload;
    function ReadCryptHeader(fSource : TStream) : integer; overload;
    function DecryptFile : integer;
    procedure EndThread (Error : integer); virtual;
    procedure UpdateProgress;
    procedure DoProgress (AAction : TFileAction; ACount : int64; ASpeedLimit : boolean = false); virtual;
    function OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
    function OpenFileForRead (const Filename : string) : integer;
    function CloseFiles (ACopyStorage,ACopyAltStreams,ACopyAcl : boolean;
                         FTime : TFileTimestamps; Attr : word) : integer;
    function CopyFile : integer;
    function ShCopyFile : integer;
    function GetZipSizes (fSource : TStream; var CSize,USize : int64) : boolean;
    function UnZipStream (fSource,fDest : TStream) : integer; overload;
    function UnZipStream (fSource,fDest : TStream; var ModTime : cardinal) : integer; overload;
    function VerifyFile : integer;
    function VerifyEncFile : integer;
    function CopyFileStorage : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest. file
    constructor Create (const ASourceName,ADestName : string;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    destructor Destroy; override;
    procedure CancelThread; virtual;
    procedure ReleaseStreams;
    property CheckTime : boolean read FCheckTime write FCheckTime;
    property CopyAttr: boolean read FCopyAttr write FCopyAttr;
    property CopyCreationTime : boolean read FCopyCrTm write FCopyCrTm;
    property CopyPermissions : boolean read FCopyAcl write FCopyAcl;
    property CopyStorage : boolean read FCopyStorage write FCopyStorage;
    property CopyAltStreams : boolean read FCopyAltStreams write FCopyAltStreams;
    property CopyTimeStamp : boolean read FCopyTimeStamp write FCopyTimeStamp;
    property VerifyDelay : cardinal read FMaxWait write FMaxWait;
    property Done  : boolean read GetDone;
    property ErrorType : integer read FErrFlag;
    property ErrorCode : integer read FError;
    property ShellCopy : TShellCopyModes read FShellCopy write FShellCopy;
    property OnProgress : TFileProgressEvent read FOnProgress write SetProgress;
    property Overwrite : boolean read FOverwrite write FOverwrite;
//    property SysError : cardinal read FSysError;
    property Ratio : integer read FRatio;  // compression ratio
    property SysErrorMsg : string read FSysErrMsg;
    property TempDir : string read fTmpDir write SetTempDirectory;
    property NoWriteCache : boolean write SetNoWriteCache;
    end;

  // Copy file with opt. encryption
  TEncryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    end;

  // Decrypt file
  TDecryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    end;

  // gzip compatible compression
  TGZipThread = class (TCopyThread)
  protected
    FName : string;
    function GzWriteHeader (fDest : TStream; fName : string;
                            fTime : cardinal; fEncMode : integer) : integer;
    function GzWriteTrailer (fDest : TStream; ACrc : cardinal) : integer;
    function CompressFile : integer;
    function VerifyGzFile : integer;
    function ProcessStream (sSource : TStream; ACompress : boolean) : integer;
    function CompressAsZip (ACompress : boolean) : integer;
    function VerifyFromZip : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = gz-file,
  // AEncryption = encrypting object (see AesLib), set to nil for plain compressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    property CompressionLevel : TZCompressionLevel read FCompLevel write FCompLevel;
    property CompressionType : TZStrategy read FCompType write FCompType;
    property Crc : cardinal read FCrc;
    property Filename : string read FName write FName;
    end;

  // copy file with opt. compression and/or encryption
  TWinCopyThread = class (TGZipThread)
  private
    FZipCompress : boolean;
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption  = encrypting object (see AesLib), set to nil for no encryption
  // AOptions = set of (coCompress,coZipCompress,coVerify,coOverwrite,coStorage,coAltStreams,
  //              coPermissions,coAttributes,coTimeStamp,coCreationTime,coCheckTime)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        AOptions : TCopyOptions;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    end;

  // gzip compatible uncompression
  TGUnZipThread = class (TCopyThread)
  protected
    function GzReadHeader (fSource : TStream; var fName : string;
                           var fTime : cardinal; var fEncMode  : integer;
                           var CSize,USize : int64) : integer;
    function GzSkipHeader (fSource : TStream) : integer;
    function GzReadSizes (fSource : TStream; var fEncMode    : integer;
                          var CSize,USize : int64) : integer;
    function GzReadTrailer (fSource : TStream; USize : int64) : integer;
    function UnCompressFile : integer;
    procedure Execute; override;
  public
  // ASourceName  = gz-file,  ADestFolder = destination folder
  // ADecryption = Decrypting object (see AesLib), set to nil for plain uncompressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    property Crc : cardinal read FCrc;
    end;

  TUnZipThread = class (TCopyThread)
  protected
    function UnZipFile : integer;
    procedure Execute; override;
  public
  // ASourceName  = gz-file,  ADestFolder = destination folder
  // ADecryption = Decrypting object (see AesLib), set to nil for plain uncompressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    property Crc : cardinal read FCrc;
    end;

  // compare files
  TCompareThread = class (TGUnZipThread)
  private
    FCompress,FZipCompress : boolean;
  protected
    function CompareDecryptedFile (fSource,fDest : TStream) : integer;
    function CompareCompressedFile(fSource,fDest : TStream) : integer;
    function CompareZipFile(fSource,fDest : TStream) : integer;
    procedure Execute; override;
  public
    constructor Create (const AVerifyName,ARefName : string;  AEncryption : TEncryption;
                        ACompress,AZipCompress : Boolean;
                        ASuspend : Boolean = false; ABufSize : integer = defBufferSize;
                        APriority : TThreadPriority = tpNormal);
    end;

  // Zip compatible compression with optional encryption
  TZipWriteThread = class (TGZipThread)
  protected
    function AddToZip : integer;
    procedure Execute; override;
  public
    constructor Create (const AFileName : string; AEncryption : TEncryption;
                        ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent); overload;
    constructor Create (ASource : TStream; AEncryption : TEncryption;
                        ZStream : TStream; ACompress,ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent); overload;
    end;

  // Zip compatible decompression with optional encryption for verification
  TZipVerifyThread = class (TGZipThread)
  private
    FLength : int64;
  protected
    function VerifyZippedFile : integer;
    procedure Execute; override;
  public
    constructor Create (AEncryption : TEncryption; ACrc : cardinal; ZStream : TStream;
                        ALength : int64; ACompress,ASuspend,ASpeedLimit : Boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent);
    end;

  // Zip compatible decompression with optional encryption
  // write to file or stream
  TZipReadThread = class (TGUnZipThread)
  private
    FZipInfo : TZipFileInfo;    // data from local header
    RdLocalHeader : boolean;
  protected
    function CopyToStream (const AZipInfo : TZipFileInfo; fDest : TStream) : integer;
    function CopyToFile : integer;
    procedure Execute; override;
  public
    constructor Create (ZStream : TStream; const AFileName : string;
                        AEncryption : TEncryption; AZipInfo : TZipFileInfo;
                        ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent); overload;
    constructor Create (ZStream,ADest : TStream;
                        AEncryption : TEncryption; AZipInfo : TZipFileInfo;
                        ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent); overload;
    constructor Create (ZStream,ADest : TStream;
                        ASuspend : Boolean; ASpeedLimit : boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TFileProgressEvent); overload;
    end;

  TZipCompareThread = class (TZipReadThread)
  protected
    function CompareToStream(const AZipInfo : TZipFileInfo; fDest : TStream) : integer;
    function CompareToFile : integer;
    procedure Execute; override;
    end;

  TZip = class (TObject)
  private
    FBufferSize : integer;
    FIsTemp,
    FEncrypt,
    FErrorFlag  : boolean;
    FEncryption : TEncryption;
    PkExtraCrypt : TPKExtraEncField;
    FOnProgress : TFileProgressEvent;
    FOnTimeout : TTimeOutEvent;
    procedure SetTempDirectory (const ATempDir : string);
    procedure SetProgress (ACallBack : TFileProgressEvent);
    function GetEmpty : boolean;
    function GetPosition : int64;
    procedure WritePosition (Value : int64);
  protected
    FStream     : TFileStream;
    FPriority   : TThreadPriority;
    fZipTmpDir,
    FBaseDir,ZName : string;
    FNumSeg,
    FCurSeg     : integer;
    FileList    : TStringList;
    procedure DoProgress(AAction : TFileAction; ACount : int64; ASpeedLimit : boolean);
    function OpenNextSegment (var nSeg : word; var sSource : TFileStream): integer;
    function OpenSegment (nSeg : word; var sSource : TFileStream): integer;
    function CopyStream (fSource,fDest : TStream; SLength : int64;
                         var nTotal : int64; var ErrMsg : string) : integer;
    property TempDir : string read fZipTmpDir write SetTempDirectory;
  public
    constructor Create (const ZipName : string; const Password : AnsiString;
                        AKeyMode,ABufSize,AMode : integer);
    destructor Destroy; override;
    procedure InitEncryption (const Password : AnsiString; AKeyMode : integer);
    procedure CopyToZip(AZip : TZip; SLength : int64);
    property OnProgress : TFileProgressEvent read FOnProgress write SetProgress;
    property OnTimeout : TTimeOutEvent read FOnTimeout write FOnTimeout;
    property ErrorFlag : boolean read FErrorFlag write FErrorFlag;
    property IsEmpty : boolean read GetEmpty;
    property IsTemp : boolean read FIsTemp write FIsTemp;
    property FilePosition : int64 read GetPosition write WritePosition;
    property NumberOfSegments : integer read FNumSeg;
    property ThreadPriority : TThreadPriority read FPriority write FPriority;
    end;

  TZipRead = class (TZip)
  private
    ZipThread   : TGUnZipThread;
    bod : int64;
  protected
    function ReadLocalHeader (sSource : TStream; AOfs : int64; var ZipInfo : TZipFileInfo) : boolean;
    function ProcessRead (sSource : TStream; sDest : TFileStream;
                          AZipInfo : TZipFileInfo; var ErrMsg : string) : integer;
    function ProcessCompare (sSource,sDest : TStream; AZipInfo : TZipFileInfo;
                             var ErrMsg : string) : integer;
    function ExtractFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
    function ExtractSplittedFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
    function CompareFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
    function CompareSplittedFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
  public
    constructor Create (const ZipName : string; const Password : AnsiString;
                        AKeyMode : integer = 0; ABufSize : integer = defBufferSize);
    destructor Destroy; override;
    procedure ReadDirectory;
    function Filename (Index : integer) :string;
    function IsEncrypted (Index : integer) : boolean;
    procedure ExtractToFile (pkHeader : TPkHeader; const DestFilename : string); overload;
//    procedure ExtractToFile (Offset : int64; TimeStamp,Attr : integer;
//                             const DestFilename : string; Segment : word =0); overload;
    procedure ExtractToFile (Index : integer; const DestFilename : string); overload;
    function ExtractToStream (const Filename : string; sDest : TStream) : boolean;
    procedure CompareToFile (pkHeader : TPkHeader; const RefFilename : string); overload;
    procedure CompareToFile (Index : integer; const DestFilename : string); overload;
    procedure Cancel;
    property Directory : TStringList read FileList;
    property DirectoryOffset : int64 read bod;
    end;

  TZipWrite = class (TZip)
  private
    FOnVerifyProgress : TFileProgressEvent;
    FOnFullDisk       : TConfirmEvent;
    ZipThread         : TGZipThread;
    fMaxSize          : int64;
    procedure SkipLocalHeader(sSource : TStream; AOfs : int64);
    function ProcessVerify (sSource : TStream; PkHeader : TPkHeader; var ErrMsg : string) : integer;
    function VerifyFile (PkHeader : TPkHeader; var ErrMsg : string) : integer;
    function VerifySplittedFile(const ZName : string;
                        PkHeader : TPkHeader; var ErrMsg : string) : integer;
  public
    constructor Create (const ZipName,BasicDirectory : string; const Password : AnsiString = '';
                        AKeyMode : integer = 0; ABufSize : integer = defBufferSize); overload;
    constructor Create (const ZipName,BasicDirectory : string; AMaxSize : int64;
                        const Password : AnsiString = '';
                        AKeyMode : integer = 0; ABufSize : integer = defBufferSize); overload;
    destructor Destroy; override;
    function NewSegment : integer;
    function CheckSegment (ASpace: Longint): boolean;
    procedure AddDir (DirName : string);
    function RemLastDir : string;
    procedure AddFile (const SourceName,FileName : string; ACompress : boolean = true); overload;
    procedure AddStream (ASource : TStream; const AName : string; ACompress : boolean = true);
    procedure AddFile (const SourceName : string; var ARatio : Integer; ACompress : boolean = true); overload;
    procedure AddFile (const SourceName : string; ACompress : boolean = true); overload;
    procedure AddFile (const SourceName : string; FileData  : TWin32FindData;ACompress : boolean = true); overload;
    procedure AddFile (const SourceName,FileName : string; var ARatio : Integer; ACompress : boolean = true); overload;
    procedure AddFile (const SourceName,FileName : string; FileData  : TWin32FindData;
                       var ARatio : Integer; ACompress : boolean = true); overload;
    procedure AddFileFrom (const SourceName,FilePath : string; FileData : TWin32FindData;
                           var ARatio : Integer; ACompress : boolean = true);
    procedure WriteDirectory;
    procedure DeleteZipFile;
    procedure RenameZipFile (const NewName : string);
    property MaxSize : int64 read fMaxSize;
    property Directory : TStringList read FileList write FileList;
    procedure Verify;
    procedure Suspend;
    procedure Resume;
    procedure Cancel;
    property TempDir;
    property OnVerifyProgress : TFileProgressEvent read FOnVerifyProgress write FOnVerifyProgress;
    property OnFullDisk : TConfirmEvent read FOnFullDisk write FOnFullDisk;
    end;

// get error message
function GetCopyErrMsg (AError : integer) : string;

// retrieve gz file info
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
function GzFileInfoXL (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
function GzTrailer (const Filename: string; var CRC,USize : cardinal) : boolean;

// retrieve info from enc file
function GetEncHeader (fSource : TStream; var KeyMode : byte; var EncHd : TEncHeader) : boolean;
function ReadEncInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;

// retrieve info from single file zip
function ZipFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : TZipState;
function GetZipLocalHeader (fSource : TStream; var ZipInfo : TZipFileInfo) : boolean;
function LastZipSegment (const Filename: string) : integer;
function GetSegmentName (const Filename : string; n : integer) : string;
function DeleteZipArchive (const Filename: string) : integer;
function CopyZipArchive (const OldName,NewName : string) : integer;
function RenameZipArchive (const OldName,NewName : string; MaxWait : integer = 0) : integer;

// check file extension for zip, gz or enc
function IsZipFile (const Filename : string) : boolean;
function IsGzFile (const Filename : string) : boolean;
function IsGzeFile (const Filename : string) : boolean;
function IsEncFile (const Filename : string) : boolean;

//* Umwandeln eines OEM-Strings nach ANSI
function StrToAnsi (const s : AnsiString) : AnsiString;

implementation

uses Vcl.Forms, System.RtlConsts, Web.HTTPApp, XlFileUtils, FilePropUtils,
  WinApiUtils, WinShell, FileConsts;

const
  Z32MaxSize = $FF000000;     // max. file size for zip without z64
  defMaxWait = 5;          // max wait time for file open before verify

{ ---------------------------------------------------------------- }
// convert from ISO-8859-1 (ANSI) to Code page 437 (OEM)
// characters not found in the conversion table are to $20 = Space
type
  TCvTable = array [0..127] of AnsiChar;

  TWord = record
    case integer of
    0 : (Both : word);
    1 : (Lo,Hi : byte);
    end;

  TLongWord = record
    case integer of
    0 : (LongWord : cardinal);
    1 : (Lo,Hi : word);
    end;

const
  TabToOEM : TCvTable =(
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$80: control char.
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$90
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$AD,#$9B,#$9C,#$20,#$9D,#$20,#$15,   //$A0
    #$20,#$20,#$20,#$AE,#$AA,#$20,#$20,#$20,
    #$F8,#$F1,#$FD,#$20,#$20,#$E6,#$14,#$20,   //$B0
    #$20,#$20,#$20,#$AF,#$AC,#$AB,#$20,#$A8,
    #$20,#$20,#$20,#$20,#$8E,#$8F,#$92,#$80,   //$C0
    #$20,#$90,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$A5,#$20,#$20,#$20,#$20,#$99,#$20,   //$D0
    #$20,#$20,#$20,#$20,#$9A,#$20,#$20,#$E1,
    #$85,#$20,#$83,#$20,#$84,#$86,#$91,#$87,   //$E0
    #$8A,#$82,#$88,#$89,#$8D,#$A1,#$8C,#$8B,
    #$20,#$A4,#$95,#$A2,#$93,#$20,#$94,#$F6,   //$F0
    #$20,#$97,#$A3,#$96,#$81,#$20,#$20,#$98);

  FILE_WRITE_ATTRIBUTES = $0100;

// Check if this Unicode string can be converted to OEM
function IsOEMStr (ws : string) : boolean;
var
  i : integer;
begin
  Result:=false;
  // check for Unicode characters
  for i:=1 to length(ws) do if TWord(ws[i]).Hi<>0 then exit;
  // check for ANSI characters not supported in code page 437
  for i:=1 to length(ws) do
    if (TWord(ws[i]).Lo>=$80) and (TabToOEM[TWord(ws[i]).Lo-128]=#$20) then exit;
  Result:=true;
  end;

// convert ANSI string to OEM - required for Zip
function StrToOEM (s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  for i:=1 to length(s) do begin
    if s[i]>=#128 then s[i]:=TabToOEM[ord(s[i])-128];
    end;
  Result:=s;
  end;

{ ---------------------------------------------------------------- }
// convert from Code page 437 (OEM) to ISO-8859-1 (ANSI)
const
// characters not found in the conversion table are set to $D7 = multiplication sign
  TabToANSI : TCvTable =(
    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,   //$80
    #$EA,#$EB,#$E8,#$EF,#$EE,#$EC,#$C4,#$C5,
    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,   //$90
    #$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$D7,#$D7,
    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$D7,#$D7,   //$A0
    #$BF,#$D7,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
    #$D7,#$D7,#$D7,#$A6,#$D7,#$D7,#$D7,#$D7,   //$B0
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,   //$C0
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,   //$D0
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,
    #$D7,#$DF,#$D7,#$D7,#$D7,#$D7,#$B5,#$D7,   //$E0
    #$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,#$D7,
    #$D7,#$B1,#$D7,#$D7,#$D7,#$D7,#$F7,#$D7,   //$F0
    #$B0,#$D7,#$B7,#$D7,#$D7,#$B2,#$D7,#$A0);

// characters not found in the conversion table are set to $20 = Space
//  TabToANSI : TCvTable =(
//    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,   //$80
//    #$EA,#$EB,#$E8,#$EF,#$EE,#$EC,#$C4,#$C5,
//    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,   //$90
//    #$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$20,#$20,
//    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$20,#$20,   //$A0
//    #$BF,#$20,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
//    #$20,#$20,#$20,#$A6,#$20,#$20,#$20,#$20,   //$B0
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$C0
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,   //$D0
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
//    #$20,#$DF,#$20,#$20,#$20,#$20,#$B5,#$20,   //$E0
//    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
//    #$20,#$B1,#$20,#$20,#$20,#$20,#$F7,#$20,   //$F0
//    #$B0,#$20,#$B7,#$20,#$20,#$B2,#$20,#$A0);

(* Umwandeln eines OEM-Strings nach ANSI *)
function StrToAnsi (const s : AnsiString) : AnsiString;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToANSI[ord(Result[i])-128]
    else if Result[i]=#$15 then Result[i]:=#$A7
    else if Result[i]=#$14 then Result[i]:=#$B6;
    end;
  end;

// cardinal from H- und Lo word
function WordsToCardinal (HiWord,LoWord : word) : cardinal;
var
  Long : TLongWord;
begin
  with Long do begin
    Hi:=HiWord; Lo:=LoWord;
    Result:=LongWord;
    end;
  end;

function GetLastErrMsg : string;
begin
  Result:=Format(rsErrSystem,[SysErrorMessage(GetLastError)]);
  end;

{ ------------------------------------------------------------------- }
function GetCopyErrMsg (AError : integer) : string;
var
  s : string;
begin
  case AError and errAllTypes of
  errError   : s:=rsError;
  errError2  : s:=rsError2;
  errWarning : s:=rsWarning;
  else s:=''; //rsInfo;
    end;
  case AError and errAllCodes of
  errFileCreate : s:=s+rsFileCreate;
  errFileOpen   : s:=s+rsFileOpen; //Could not open file
  errFileClose  : s:=s+rsFileClose; //Could not close file
  errFileRead   : s:=s+rsFileRead; //Could not read from file
  errFileWrite  : s:=s+rsFileWrite;  // Could not write to file
  errFileAttr   : s:=s+rsFileAttr;   // File attributes could not be set
  errFileFull   : s:=s+rsFileFull;   // Low disk space
  errFileGZip   : s:=s+rsFileGZip;   // Illegal file header
  errFileCheck  : s:=s+rsFileCheck;  // Corrupt file
  errFileECrypt : s:=s+rsFileECrypt; // Encryption failed
  errFileDCrypt : s:=s+rsFileDCrypt; // Decryption failed
  errFileVerify : s:=s+rsFileVerify; // Verification failed
  errLongPath   : s:=s+rsLongPath;   // Path too long
  errNotFound   : s:=s+rsNotFound;   // File not found
  errFileTS     : s:=s+rsFileTS;     // Timestamp could not be set
  errStorage    : s:=s+rsStorage;    // Error copying document summary
  errTimeout    : s:=s+rsTimeout;    // Timeout error on copying file
  errStream     : s:=s+rsStream;     // Undefined stream
  errAcl        : s:=s+rsAcl;        // Permissions could not be copied
  errFileExists : s:=s+rsFileExists; // File already exists
  errSzMismatch : s:=s+rsSzMismatch; // Size mismatch
  errVerOpen    : s:=s+rsVerOpen;    // Could not open destination file for verification
  errDirCreate  : s:=s+rsDirCreate;  // Ccould not create directory
  errFtpRead    : s:=s+rsFtpRead;    // Could not read from FTP
  errFtpWrite   : s:=s+rsFtpWrite;   // Could not write to FTP
  errFtpConnect : s:=s+rsFtpConnect; // Could not connect to FTP server
  errFtpBroken  : s:=s+rsFtpBroken;  // FTP connection was closed by server
  errFtpDatConn : s:=s+rsFtpDatConn; // Could not open FTP data connection
  errFtpTimeout : s:=s+rsFtpTimeout; // Timeout error on copying file via FTP
  errCompare    : s:=s+rsCompare;    // Contents mismatch
  errZipCrSeg   : s:=s+rsZipCrSeg;   // Error creating new zip segment
  errSignature  : s:=s+rsSignature;  // Illegal zip signature
  errExtract    : s:=s+rsExtract;    // Could not extract file
  errFormat     : s:=s+rsFormat;     // Unsupported file format
  errTmpFile    : s:=s+rsTmpFile;    // Could not rename temporary file
  errZipRdSeg   : s:=s+rsZipRdSeg;   // Error reading next zip segment
  errAltStreams : s:=s+rsAltStreams;  // Error copying alternate file streams

  errUserBreak  : s:=s+rsUserBreak; //Terminated by user
  else s:=s+Format(rsUnknownErrCode,[AError]); // should not happen
    end;
  case AError and errAllSources of
  errCopy       : s:=s+rsCopy; // (Copy)
  errGZip       : s:=s+rsGZip; // (gzip)
  errGUnzip     : s:=s+rsGUnzip; // (gunzip)
  errZip        : s:=s+rsZip; // (Zip)
  errUnzip      : s:=s+rsUnzip; // (Unzip)
  errEnCrypt    : s:=s+rsEnCrypt; // (Encrypt)
  errDeCrypt    : s:=s+rsDeCrypt; // (Decrypt)
    end;
  Result:=s;
  end;

  { ------------------------------------------------------------------- }
function GetGzHeader (fSource   : TStream;
                      var fName : string;
                      var fTime : cardinal;
                      var fEncMode    : integer;
                      var CSize,USize : int64) : integer;
var
  Header : TGzHeader;
  n,k,m  : word;
  b      : byte;
  s      : RawByteString;
  fp     : int64;
  Buffer : array of byte;
begin
  Result:=errOK;       // o.k.
  fName:=''; fTime:=0;
  try
    fSource.Read(Header,sizeof(TGzHeader));
    with Header do begin
      if Signature<>GzSignatur then begin
        Result:=errFileGZip; exit;
        end;
      fTime:=TimeStamp;
      with fSource do begin
        CSize:=0; USize:=0; m:=0;
        if Flags and gfExtraField <> 0 then begin   // read extra field
          Read(n,2);              // length of extra field
          while n>sizeof(word) do begin
            Read(k,sizeof(word));     // ID
            dec(n,2);
            if k=ExtraGz64IDA then begin
              Read(k,sizeof(word));  // Data size (= 16)
              Read(CSize,sizeof(int64));
              Read(USize,sizeof(int64));
              dec(n,2+2*sizeof(int64));
              end
            else if k=ExtraGzCrypt then begin
              Read(k,sizeof(word));  // Data size (= 2)
              Read(m,sizeof(word));
              dec(n,2+sizeof(word));
              end
            else begin    // overread
              Read(k,sizeof(word));  // Data size
              SetLength(Buffer,k);
              Read(Buffer[0],k);
              dec(n,2+k);
              end;
            end;
          Position:=Position+n;   // adjust stream position
          end;
        if Flags and gfFileName <> 0 then begin   // read filename
          repeat
            Read(b,1);
            if b<>0 then s:=s+AnsiChar(b);
            until b=0;
          end;
        if Flags and gfComment <> 0 then begin   // skip comment
          repeat
            Read(b,1);
            until b=0;
          end;
        if Flags and gfHeaderCrc <> 0 then Read(n,2); // skip crc16
        if Flags and gfEncrypted<>0 then fEncMode:=m else fEncMode:=-1;
        if OpSys=OsDos then fName:=s
        else fName:=Utf8ToString(s);
        if CSize=0 then CSize:=Size-Position-8;
        if USize=0 then begin
          fp:=Position;
          Seek (-4,soFromEnd);
          Read(USize,4);
          Position:=fp;
          end;
        end;
      end;
  except
    Result:=errFileRead; // error
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from stream *)
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
begin
  with FileInfo do begin
    Count:=1; Attr:=InvalidAttributes;
    UnixTime:=DateTimeToUnixTime(Now);
    if GetGzHeader (sGz,Filename,UnixTime,EncMode,CSize,USize)=errOK then begin
      Encrypted:=EncMode>=0;
      Result:=true
      end
    else Result:=false;
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from gz-file *)
function GzFileInfoXL (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
var
  sFile : TFileStream;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(FilenameToXL(Filename),fmOpenRead+fmShareDenyNone);
    Result:=GetGzInfo(sFile,FileInfo);
    if Result then FileInfo.Attr:=FileGetAttrXL(Filename,false);
  finally
    sFile.Free;
    end;
  end;

function GzTrailer (const Filename: string; var CRC,USize : cardinal) : boolean;
var
  sFile  : TFileStream;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(FilenameToXL(Filename),fmOpenRead+fmShareDenyNone);
    with sFile do begin
      Seek(-8,soEnd);
      Read(Crc,4);
      Read(USize,4);
      end;
    Result:=true;
  finally
    sFile.Free;
    end;
  end;

{ ------------------------------------------------------------------- }
// read header with signature from encrypted file to get keymode
function GetEncHeader (fSource : TStream; var KeyMode : byte; var EncHd : TEncHeader) : boolean;
var
  hh : boolean;
begin
  Result:=true;
  with EncHd do begin
    TimeStamp:=0; Attr:=InvalidAttributes;
    end;
  try
    fSource.Read(EncHd.Signature,sizeof(cardinal));
    KeyMode:=(LongRec(EncHd.Signature).Bytes[3] shr 3) and 3;   // Keymode;
    hh:=LongRec(EncHd.Signature).Bytes[3] and 1 <>0; // extended header
    if EncHd.Signature and $E6FFFFFF <>CrSignatur then begin
      KeyMode:=0; EncHd.Signature:=0;
      fSource.Position:=0; // no signature found, former data format
      end
    else begin
      if hh then with fSource do begin
        Position:=0;
        Read(EncHd,sizeof(TEncHeader));
        end;
      end;
  except
    Result:=false; // error
    end;
  end;

function ReadEncInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
var
  sFile : TFileStream;
  eh    : TEncHeader;
  km    : byte;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(FilenameToXL(Filename),fmOpenRead+fmShareDenyNone);
    Result:=GetEncHeader(sFile,km,eh);
    if Result then with FileInfo do begin
      UnixTime:=eh.TimeStamp; Attr:=eh.Attr;
      end;
  finally
    sFile.Free;
    end;
  end;

{ ------------------------------------------------------------------- }
function GetZipLocalHeader (fSource : TStream; var ZipInfo : TZipFileInfo) : boolean;
var
  LFHeader  : TPKLocalHeader;
  el,k      : cardinal;
  pkelf     : TExtraFieldInfo;
  Buffer    : array of byte;
  ZfName    : RawByteString;
begin
  try
    fSource.Read(LFHeader,sizeof(TPKLocalHeader));         // local header
    with LFHeader do if Signatur=PKLocalSignatur then begin
      ZipInfo.TimeStamp:=FTimeStamp;
      ZipInfo.Crc:=Crc;
      ZipInfo.CSize:=CSize;
      ZipInfo.USize:=USize;
      ZipInfo.Flag:=Flag;
      ZipInfo.Method:=Method;
      ZipInfo.KeyMode:=1;      // default
      SetLength(ZfName,FNLength);
      fSource.Read(ZfName[1],FNLength);
      if Flag and zipUtf8Flag<>0 then ZipInfo.Filename:=UnixPathToDosPath(UTF8ToWideString(ZfName))
      else ZipInfo.Filename:=UnixPathToDosPath(StrToAnsi(ZfName));
      if ExtraLength>0 then begin      // read extra fields
        el:=ExtraLength;
        while el>0 do begin  // read extra fields
          fSource.Read(pkelf,SizeOf(TExtraFieldInfo));
          with pkelf do begin
            SetLength(Buffer,ESize);
            fSource.Read(Buffer[0],ESize);
            k:=0;
            if Tag=ExtraZip64 then begin
              if USize=MaxCardinal then begin
                Move(Buffer[k],ZipInfo.USize,8);
                inc(k,8);
                end;
              if (CSize=MaxCardinal) and (ESize>k) then begin
                Move(Buffer[k],ZipInfo.CSize,8);
                end
              end
            else if Tag=ExtraCrypt then begin
              ZipInfo.KeyMode:=Buffer[4];         // Mode
              end;
            el:=el-SizeOf(TExtraFieldInfo)-ESize;
            end;
          end;
        Buffer:=nil;
        end;
      Result:=true;
      end
    else Result:=false;
  except
    Result:=false;
    end;
  end;

// retrieve number of entries in zip file
function GetZipEntries (fSource : TStream) : int64;
var
  pke   : TPkEndHeader;
  pkz   : TPKZip64CdRecord;
  pko   : TPKZip64CdLocator;
begin
  Result:=-1;
  with fSource do begin
    Position:=Size-sizeof(pke);
    Read(pke,sizeof(pke));
    end;
  with pke do begin
    if Signatur<>PkEndSignatur then Exit;
    if (pke.Offset=MaxCardinal) or (pke.DirSize=MaxCardinal) or (pke.TotalEntries=MaxWord) then begin // Zip64 extension
      with fSource do begin
        Position:=Size-sizeof(pke)-sizeof(pko);  // read Zip64 end of central directory locator
        Read(pko,sizeof(pko));
        end;
      if pko.Signatur<>Pk64LocSignatur then exit;
      fSource.Position:=pko.Offset;                    // read Zip64 end of central directory record
      fSource.Read(pkz,sizeof(pkz));
      if pkz.Signatur<>Pk64RecSignatur then Exit;
      Result:=pkz.TotalEntries;
      end
    else Result:=pke.TotalEntries;
    end;
  fSource.Position:=0;
  end;

// retrieve info from first entry of zip file
// Result = zsSingle : zip file with 1 entry
// Result = zsMulti : zip file with more than 1 entries
// Result = zsError : error reading zip info
function ZipFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : TZipState;
var
  sFile : TFileStream;
  ZipInfo : TZipFileInfo;
begin
  Result:=zsError;
  try
    sFile:= TFileStream.Create(FilenameToXL(Filename),fmOpenRead+fmShareDenyNone);
    with FileInfo do begin
      UnixTime:=DateTimeToUnixTime(Now);
      Count:=GetZipEntries(sFile);
      if Count>0 then begin
        if GetZipLocalHeader (sFile,ZipInfo) then begin
          Filename:=ZipInfo.Filename;
          UnixTime:=FileAgeToUnixTime(ZipInfo.TimeStamp);
          Encrypted:=ZipInfo.Flag and zipEncryptFlag<>0;
          CSize:=ZipInfo.CSize;
          USize:=ZipInfo.USize;
          if (Count=1) and ((ZipInfo.Flag and zipPbFlag)<>0) then Result:=zsSingle
          else Result:=zsMulti;
          end;
        end;
      Attr:=FileGetAttrXL(Filename,false);   // local header has no attribute
      end;
  finally
    sFile.Free;
    end;
  end;

// get number of segments of zip archive
function LastZipSegment (const Filename: string) : integer;
var
  sFile : TFileStream;
  pke   : TPkEndHeader;
  pko   : TPKZip64CdLocator;
begin
  Result:=0;
  try
    sFile:= TFileStream.Create(FilenameToXL(Filename),fmOpenRead+fmShareDenyNone);
    with sFile do begin
      Position:=Size-sizeof(pke);
      Read(pke,sizeof(pke));
      end;
    with pke do begin
      if Signatur<>PkEndSignatur then Exit;
      if (pke.Offset=MaxCardinal) or (pke.DirSize=MaxCardinal) or (pke.TotalEntries=MaxWord) then begin // Zip64 extension
        with sFile do begin
          Position:=Size-sizeof(pke)-sizeof(pko);  // read Zip64 end of central directory locator
          Read(pko,sizeof(pko));
          end;
        if pko.Signatur<>Pk64LocSignatur then exit;
        Result:=pko.TotalDisk-1;
        end
      else Result:=pke.ThisDisk;
      end;
  finally
    sFile.Free;
    end;
  end;

function GetSegmentName (const Filename : string; n : integer) : string;
begin
  Result:=FilenameToXL(NewExt(Filename,'z'+Format('%.2d',[n+1])));
  end;

function DeleteZipArchive (const Filename: string) : integer;
var
  n,i : cardinal;
begin
  if FileExistsXL(Filename) then begin
    n:=LastZipSegment(FileName);
    if n>0 then begin // splitted
      for i:=0 to n-1 do begin
        Result:=DeleteExistingFileXL(GetSegmentName(Filename,i));
        if Result<>NO_ERROR then Exit;
        end;
      end;
    Result:=DeleteExistingFileXL(Filename);
    end
  else Result:=NO_ERROR;
  end;

function CopyZipArchive (const OldName,NewName : string) : integer;
var
  n,i : cardinal;
begin
  if FileExistsXL(OldName) then begin
    Result:=ERROR_SUCCESS;
    n:=LastZipSegment(OldName);
    if n>0 then begin // splitted
      for i:=0 to n-1 do begin
        try
          CopyFileTS(GetSegmentName(OldName,i),GetSegmentName(NewName,i));
        except
          on E:ECopyError do begin
            Result:=E.ErrorCode;
            Exit;
            end;
          end;
        end;
      end;
    try
      CopyFileTS(OldName,NewName);
    except
      on E:ECopyError do begin
        Result:=E.ErrorCode;
        Exit;
        end;
      end;
    end
  else Result:=ERROR_FILE_NOT_FOUND;
  end;

function RenameZipArchive (const OldName,NewName : string; MaxWait : integer) : integer;
var
  n,i,k,del : cardinal;

  function TryRenameFile (const OldName,NewName : string) : integer;
  begin
    if del=0 then Result:=RenameFileXL(OldName,NewName)
    else repeat
      Result:=RenameFileXL(OldName,NewName);
      if Result<>NO_ERROR then sleep(del);  // try again
      inc(k);
      until (Result=NO_ERROR) or (k>25); // 5 s
    end;

begin
  k:=0;
  if MaxWait=0 then del:=0
  else del:=1000*MaxWait div 25;  // try 25 times until time out
  if FileExistsXL(OldName) then begin
    n:=LastZipSegment(OldName);
    if n>0 then begin // splitted
      for i:=0 to n-1 do begin
        Result:=TryRenameFile(GetSegmentName(OldName,i),GetSegmentName(NewName,i));
        if Result<>NO_ERROR then Exit;
        end;
      end;
    Result:=TryRenameFile(OldName,NewName);
    end
  else Result:=ERROR_FILE_NOT_FOUND;
  end;

{ ------------------------------------------------------------------- }
// check file types
function IsGzFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=GzExt;
  end;

function IsGzeFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=GzeExt;
  end;

function IsEncFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=CrExt;
  end;

function IsZipFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=ZipExt;
  end;

{ ------------------------------------------------------------------- }
function ZipCrc32(Crc : cardinal; const buf; len: Integer): cardinal;
begin
  Result:=Crc32 (Crc,pbyte(@buf),len);
  end;

{ ------------------------------------------------------------------- }
{ EWriteFileStreamError }
constructor EWriteFileStreamError.Create (ECode : integer; const EMsg : string);
begin
  FCode:=ECode;
  inherited Create(EMsg);
  end;

constructor TWriteFileStream.Create (const AFileName: string; Overwrite : boolean = true;
                                     WriteMode : cardinal = wmSequential; ShareMode: Word = 0);
var
  ec : integer;
  cd : cardinal;
begin
  if Overwrite then cd:=CREATE_ALWAYS else cd:=CREATE_NEW;
  inherited Create(CreateFile(PChar(AFileName),GENERIC_WRITE,
      ShareMode,nil,cd,FILE_ATTRIBUTE_NORMAL or WriteMode,0));
  if FHandle = INVALID_HANDLE_VALUE then begin
    ec:=GetLastError;
    raise EWriteFileStreamError.Create(ec,SysErrorMessage(ec));
    end;
  end;

function TWriteFileStream.Write(const Buffer; Count: Longint): Longint;
var
  ec : integer;
begin
  Result:=FileWrite(FHandle, Buffer, Count);
  if Result=-1 then begin
    ec:=GetLastError;
    raise EWriteFileStreamError.Create(ec,SysErrorMessage(ec));
    Result:=0;
    end;
  end;


{ ------------------------------------------------------------------- }
constructor TCopyStreamThread.Create (ASource,ADest : TStream; ALength : int64;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASuspend);
  sSource:=ASource; sDest:=ADest;
  FLength:=ALength;
  Priority:=APriority;
  FOnSProgress:=nil;
  FBufSize:=ABufSize;
  SetLength(FBuffer,FBufSize);
  FError:=errOK;
  FUserBreak:=false;
  FSysErrMsg:='';
  FTotal:=0;
  end;

destructor TCopyStreamThread.Destroy;
begin
  FBuffer:=nil;
  inherited Destroy;
  end;

function TCopyStreamThread.GetDone : boolean;
begin
  Result:=Terminated;
  end;

procedure TCopyStreamThread.SetProgress (ACallBack : TFileProgressEvent);
begin
  FOnSProgress:=ACallBack;
  end;

procedure TCopyStreamThread.DoProgress (AAction : TFileAction; ACount : int64);
begin
  if Assigned(FOnSProgress) then begin
    FAction:=AAction; FCount:=ACount;
    Synchronize(UpdateProgress);
    end;
  end;

procedure TCopyStreamThread.UpdateProgress;
begin
  FOnSProgress(FAction,FCount,false);
  end;

function TCopyStreamThread.CopyStream : integer;
var
  NRead,NWrite,
  fb            : Integer;
begin
  Result:=errOK;
  repeat
    if FLength<FBufSize then fb:=FLength else fb:=FBufSize;
    try
      NRead:=sSource.Read(FBuffer[0],fb);
      if NRead<fb then begin
        Result:=errFileRead;   // z.B. wenn "sSource" gelockt ist
        FSysErrMsg:=GetLastErrMsg;
        end;
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        Result:=errFileRead;
        end;
      end;
    if Result=errOK then begin
      inc(FTotal,NRead);
      DoProgress(acCopy,FTotal);
      try
        NWrite:=sDest.Write(FBuffer[0],NRead);
        if NWrite<NRead then Result:=errFileFull; // Ziel-Medium voll
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          Result:=errFileWrite;
          end;
        end;
      end;
    dec(Flength,NRead);
    until (FLength<=0) or (Result<>errOK);
  FBuffer:=nil;
  end;

procedure TCopyStreamThread.EndThread (Error : integer);
begin
  if Error<>errOK then FError:=errCopy or Error;
  Terminate;
  end;

procedure TCopyStreamThread.Execute;
begin
  EndThread(CopyStream);
  end;

{ ------------------------------------------------------------------- }
constructor TCopyThread.Create (const ASourceName,ADestName : string;
                                ASuspend : Boolean; ABufSize : integer;
                                APriority : TThreadPriority);
begin
  inherited Create (ASuspend);
  SourceName:=ASourceName;
  DestName:=ADestName;
  SourceNameXL:=FilenameToXL(ASourceName);
  DestNameXL:=FilenameToXL(ADestName);
  Priority:=APriority;
  FOnProgress:=nil;
  FCopyTimeStamp:=true;
  FCopyAttr:=true;
  FCheckTime:=true;
  FCopyCrTm:=false;
  FCopyStorage:=false;
  FCopyAltStreams:=false;
  FCopyAcl:=false;
  FShellCopy:=[];
  FCompress:=false;
  FVerify:=false;
  FOverwrite:=false;
  FError:=errOK; FErrFlag:=errCopy;
  FSysError:=ERROR_SUCCESS; FSysErrMsg:='';
  FMaxWait:=defMaxWait;
  FBufSize:=ABufSize;
  SetLength(FBuffer,FBufSize);
  FUserBreak:=false;
  FWriteMode:=wmSequential;
  UTotal:=0; CTotal:=0; FRatio:=-1;
  // init checksum
  FCrc:=Crc32(0,nil,0);
  FCompLevel:=zcDefault; FCompType:=zsDefault;
  FCrypting:=nil;
  fTmpDir:=TempDirectory;
  end;

destructor TCopyThread.Destroy;
begin
  FBuffer:=nil;
  if assigned(FCrypting) then FCrypting.OnProgress:=nil;
  inherited Destroy;
  end;

// returns true if error occured, set FSysError
function TCopyThread.CheckSysError (ASysError : cardinal) : boolean;
begin
  Result:=ASysError<>ERROR_SUCCESS;
  if Result then begin
    if (TLongWord(ASysError).Hi and $7FF)=0 then FSysError:=WordsToCardinal(FACILITY_WIN32,ASysError)
    else FSysError:=ASysError;
    end;
  end;

function TCopyThread.GetDone : boolean;
begin
  Result:=Terminated;
  end;

procedure TCopyThread.SetNoWriteCache (Value : boolean);
begin
  if Value then FWriteMode:=wmSequential or wmNoCache else FWriteMode:=wmSequential;
  end;

procedure TCopyThread.SetProgress (ACallBack : TFileProgressEvent);
begin
  FOnProgress:=ACallBack;
  if assigned(FCrypting) then FCrypting.OnProgress:=ACallBack;
  end;

procedure TCopyThread.SetTempDirectory (const ATempDir : string);
begin
  if (length(ATempDir)>0) and DirectoryExists(ATempDir) then fTmpDir:=SetDirName(ATempDir)
  else fTmpDir:=TempDirectory;
  end;

procedure TCopyThread.EndThread (Error : integer);
begin
  if FCompress and (UTotal>0) then FRatio:=round(1000*CTotal/UTotal);
  if FSysError<>NO_ERROR then FSysErrMsg:=Format(rsErrSystem,[SysErrorMessage(FSysError)]);
  if Error<>errOK then FError:=FErrFlag or Error;
  Terminate;
  end;

procedure TCopyThread.CancelThread;
begin
  FUserBreak:=true;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CopyStream (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean) : integer;
var
  NRead,NWrite,ec,fb : integer;
begin
  ec:=errOK;
  repeat
    if SLength<FBufSize then fb:=SLength else fb:=FBufSize;
    try
      NRead:=fSource.Read(FBuffer[0],fb);
      if NRead<fb then begin
        ec:=errError+errFileRead;   // z.B. wenn "sSource" gelockt ist
        FSysError:=GetLastError;
        end
      else begin
        UTotal:=UTotal+NRead;        // total number of bytes read
        DoProgress(acCopy,UTotal,ASpeedLimit);
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        end;
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileRead;
        end;
      end;
    if ec=errOK then begin
      try
        if assigned(fDest) then NWrite:=fDest.Write(FBuffer[0],NRead)
        else NWrite:=NRead;
        if NWrite<NRead then ec:=errError+errFileFull; // Ziel-Medium voll
      except
        on E:EWriteFileStreamError do begin
          FSysErrMsg:=E.Message;
          if E.ErrorCode=ERROR_DISK_FULL then ec:=errError+errFileFull // Ziel-Medium voll
          else ec:=errError+errFileWrite;
          end;
        end;
      if FUserBreak then ec:=errUserBreak;
      end;
    dec(Slength,NRead);
    until (SLength<=0) or (ec<>errOK);
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CompareStreams (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  NRead1,NRead2,fb : integer;
  CBuffer          : array of byte;
  SLength,STotal   : int64;
begin
  sLength:=fSource.Size;
  SetLength(CBuffer,FBufSize);
  STotal:=0;
  DoProgress(acVerify,-SLength,ASpeedLimit);
  if SLength<>fDest.Size then Result:=errError+errSzMismatch
  else begin
    fSource.Seek(0,soFromBeginning);
    fDest.Seek(0,soFromBeginning);
    Result:=errOK;
    repeat
      if SLength<FBufSize then fb:=SLength else fb:=FBufSize;
      try
        NRead1:=fSource.Read(FBuffer[0],fb);
        if NRead1<fb then begin
          Result:=errError+errFileRead;   // z.B. wenn "sSource" gelockt ist
          FSysError:=GetLastError;
          end
        else begin
          STotal:=STotal+NRead1;        // total number of bytes read
          DoProgress(acVerify,STotal,ASpeedLimit);
          end;
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          Result:=errError+errFileRead;
          end;
        end;
      if Result=errOK then begin
        try
          NRead2:=fDest.Read(CBuffer[0],NRead1);
          if NRead2<NRead1 then begin
            Result:=errError2+errFileRead;    // z.B. wenn "sDest" gelockt ist
            FSysError:=GetLastError;
            end;
        except
          on E:Exception do begin
            FSysErrMsg:=E.Message;
            Result:=errError2+errFileRead;
            end;
          end;
        if FUserBreak then Result:=errUserBreak;
        end;
      dec(Slength,NRead1);
      if (Result=errOK) and not CompareMem(@FBuffer[0],@CBuffer[0],NRead1) then
        Result:=errError+errCompare;
      until (SLength<=0) or (Result<>errOK);
    end;
  CBuffer:=nil;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.GzCompress (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  sComp         : TZCompressionStream;
  NRead         : cardinal;
  ec            : integer;
  fp            : int64;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible compression
  sComp:=TZCompressionStream.Create(fDest,FCompLevel,defWindowBits,defMemLevel,FCompType);
  fp:=fDest.Position;
  repeat
    try
      NRead:=fSource.Read(FBuffer[0],FBufSize);
      UTotal:=UTotal+NRead;        // total number of bytes read
      DoProgress(acCompress,UTotal,ASpeedLimit);
      FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileRead;
        end;
      end;
    if ec=errOK then begin
      try
        sComp.Write(FBuffer[0],NRead);
//   does not work with TZCompressionStream: NWrite is always equal to NRead
//        if NWrite<NRead then ec:=errError+errFileFull;  // dest disk full
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          ec:=errError+errFileWrite;
          end;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);
  if ec=errOK then sComp.Free else sComp.Break;
  if (ec=errOK) and (fSource.Size>0) and (fDest.Position<=fp) then
    ec:=errError+errFileFull;  // dest disk full
  Result:=ec;
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream; ASpeedLimit : boolean) : integer;
var
  ALength : int64;
begin
  with fSource do ALength:=Size-Position-8; // gz-format
  Result:=GzUnCompress (fSource,fDest,ALength,ASpeedLimit);
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream; SLength : int64; ASpeedLimit : boolean) : integer;
var
  sComp         : TZDecompressionStream;
  NRead,NWrite  : cardinal;
  RTotal        : int64;
  ec            : integer;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible decompression
  sComp:=TZDecompressionStream.Create(fSource,SLength,defWindowBits);
  RTotal:=0;
  repeat
    try
      NRead:=sComp.Read(FBuffer[0],FBufSize);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileRead;
        end;
      end;
    if ec=errOK then begin
      try
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        RTotal:=RTotal+NRead;
        DoProgress(acUnCompress,RTotal,ASpeedLimit);
        if assigned(fDest) then NWrite:=fDest.Write(FBuffer[0],NRead)
        else NWrite:=NRead;
        UTotal:=UTotal+NWrite;        // total number of bytes written
        if (NWrite<NRead) then ec:=errError+errFileFull;    // dest disk full
      except
        on E:EWriteFileStreamError do begin
          FSysErrMsg:=E.Message;
          if E.ErrorCode=ERROR_DISK_FULL then ec:=errError+errFileFull // Ziel-Medium voll
          else ec:=errError+errFileWrite;
          end;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);;
  sComp.Free;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.DoProgress (AAction : TFileAction; ACount : int64; ASpeedLimit : boolean);
begin
  if Assigned(FOnProgress) then begin
    FAction:=AAction; FCount:=ACount; FSpeedLimit:=ASpeedLimit;
    Synchronize(UpdateProgress);
    end;
  end;

procedure TCopyThread.UpdateProgress;
begin
  FOnProgress(FAction,FCount,FSpeedLimit);
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
begin
  Result:=errOK;
  if not FileExists(SourceNameXL,false) then Result:=errError+errNotFound;
  if Result=errOK then begin
    if (length(DestName)=0) then Result:=errError+errFileCreate;
    end;
  if Result<>errOK then exit;
  FTime:=GetFileTimestamps(SourceNameXL,false);
  Attr:=FileGetAttr(SourceNameXL,false);
  try
    sSource:=TFileStream.Create(SourceNameXL,fmOpenRead+fmShareDenyNone);
  except
    on E:EFOpenError do begin
      FSysErrMsg:=E.Message;
      Result:=errError+errFileOpen; Exit;
      end;
    end;
  // overwrite destination, reset attributes
  if FOverwrite and FileExists(DestNameXL,false) then begin
    if CheckSysError(FileSetAttr(DestNameXL,faArchive,false)) then Result:=errError+errFileAttr;
    end;
  if Result=errOK then begin
    try
      sDest:=TWriteFileStream.Create(DestNameXL,FOverwrite,FWriteMode);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        Result:=errError+errFileCreate;
        end;
      end;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.ReleaseStreams;
begin
  if assigned(sSource) then try FreeAndNil(sSource); except end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CloseFiles (ACopyStorage,ACopyAltStreams,ACopyAcl : boolean;
                                 FTime : TFileTimestamps; Attr : word) : integer;
var
  FAttr : word;
  ec    : integer;
begin
  Result:=errOK;
  if FVerify and not FlushFileBuffers(TFileStream(sDest).Handle) then begin
    FSysError:=GetLastError; Result:=errError+errFileWrite;
    end;
  ReleaseStreams;
  ec:=errOK;
  if (Result=errOK) then begin
    if ACopyAltStreams then begin
      if CheckSysError(CopyAlternateStreams(SourceNameXL,DestNameXL)) then ec:=errWarning+errAltStreams;
      end;
    if (ec=errOK) and ACopyStorage then begin
      if CheckSysError(CopyFileStorage) then ec:=errWarning+errStorage;
      end;
    if (ec=errOK) and ACopyAcl then begin
      if CheckSysError(CopyFileAcl(SourceNameXL,DestNameXL)) then ec:=errWarning+errAcl;
      end;
    end;
  if (Result=errOK) and FCopyTimeStamp then begin
    if CheckSysError(SetFileTimestamps(DestNameXL,FTime,FCheckTime,FCopyCrTm,false)) then
      Result:=errWarning+errFileTS;
    end;
  if (Result=errOK) and FCopyAttr then begin
    FAttr:=FileGetAttr(DestNameXL,false);
    if (FAttr<>Attr) and CheckSysError(FileSetAttr(DestNameXL,Attr,false)) then
      Result:=errWarning+errFileAttr;
    end;
  if (Result=errOK) and (ec<>errOK) then Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// This routine only supports pathnames < MAX_PATH (=260)
// Reason: the called function "StgOpenStorageEx" does not support extended-length paths
function TCopyThread.CopyFileStorage : integer;
var
  Summary : TFileSummary;
begin
  FSysError:=GetFileSummary(SourceName,Summary);
  if FSysError=S_OK then begin
    FSysError:=SetFileSummary(DestName,Summary);
    if FSysError=S_OK then Result:=errOK
    else Result:=errWarning+errStorage;
    end
  else Result:=errWarning+errStorage;
  end;

{ ------------------------------------------------------------------- }
// encrypt file
function TCopyThread.EncryptFile : integer;
var
  NRead    : cardinal;
  FTime    : TFileTimestamps;
  FAttr    : word;
  ec       : integer;
  cs       : cardinal;
  eh       : TEncHeader;
begin
  ec:=OpenFiles (FTime,FAttr);
  if (ec=errOK) then begin
    DoProgress(acEncrypt,-sSource.Size,false);   // Reset
    with FCrypting do begin
      with eh do begin
        Signature:=CrSignatur;
        with LongRec(Signature) do Bytes[3]:=Bytes[3]+KeyMode shl 3+1;
        TimeStamp:=FileTimeToUnixTime(FTime.LastWriteTime);
        Attr:=FAttr;
        end;
      try
        sDest.Write(eh,sizeof(TEncHeader));
      except
        on E:EWriteFileStreamError do ec:=errError+errFileWrite;
        end;
      if (ec=errOK) and WriteHeader (sDest) then begin
        repeat
          try
            NRead:=sSource.Read(FBuffer[0],FBufSize);
            UTotal:=UTotal+NRead;        // total number of bytes read
            DoProgress(acEncrypt,UTotal,true);
          except
            on E:Exception do begin
              FSysErrMsg:=E.Message;
              ec:=errError+errFileRead;
              end;
            end;
          if ec=errOK then begin
            try
              EncryptBlock (FBuffer[0],NRead);
              sDest.Write(FBuffer[0],NRead);
            except
              on E:EWriteFileStreamError do begin
                FSysErrMsg:=E.Message;
                if E.ErrorCode=ERROR_DISK_FULL then ec:=errError+errFileFull // Ziel-Medium voll
                else ec:=errError+errFileECrypt;
                end;
              end;
            if  FUserBreak then ec:=errUserBreak;
            end;
          until (NRead<FBufSize) or (ec<>errOK);
        if (ec=errOK) and not WriteTrailer (sDest) then ec:=errError+errFileECrypt;
        end
      else ec:=errError+errFileECrypt;
      end;
    if (ec=errOK) and (UTotal<>sSource.Size) then ec:=errError+errFileRead;  // z.B. wenn "sSource" gelockt ist
    if ec=errOK then ec:=CloseFiles (FCopyStorage,FCopyAltStreams,FCopyAcl,FTime,FAttr)
    else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify encrypted file
function TCopyThread.VerifyEncFile : integer;
begin
  Result:=OpenFileForRead(DestNameXL);
  if Result=errOK then begin
    DoProgress(acVerify,-UTotal,false);   // Reset progress indicator
    Result:=ReadCryptHeader(sDest);
    if (Result=errOK) then begin
      if not FCrypting.DecryptStream(sDest,nil,sDest.Size-sDest.Position,true) then
        Result:=errError+errFileVerify;
      end;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.ReadCryptHeader(fSource : TStream; var EncHd : TEncHeader) : integer;
var
  km : byte;
begin
  if GetEncHeader(fSource,km,EncHd) then begin+
    Result:=errOK; FCrypting.KeyMode:=km;
    end
  else Result:=errFileRead;
  end;

function TCopyThread.ReadCryptHeader(fSource : TStream) : integer;
var
  eh : TEncHeader;
begin
  Result:=ReadCryptHeader(fSource,eh);
  end;

// decrypt file
function TCopyThread.DecryptFile : integer;
var
  FTime  : TFileTimestamps;
  FAttr  : word;
  eh     : TEncHeader;
begin
  Result:=OpenFiles (FTime,FAttr);
  if (Result=errOK) then begin
    DoProgress(acDecrypt,-sSource.Size,true);
    Result:=ReadCryptHeader(sSource,eh);
    if (Result=errOK) then begin
      if not FCrypting.DecryptStream(sSource,sDest,sSource.Size-sSource.Position,true)
        then Result:=errError+errFileDCrypt;
      end;
    if Result=errOK then begin
      with eh do begin
        if TimeStamp<>0 then FTime.LastWriteTime:=UnixTimeToFileTime(TimeStamp);
        if Attr<>InvalidAttributes then FAttr:=Attr;
        end;
      Result:=CloseFiles(false,FCopyAltStreams,FCopyAcl,FTime,FAttr);
      end
    else ReleaseStreams;
    end
  else ReleaseStreams;
  end;

{ ------------------------------------------------------------------- }
// copy file
function TCopyThread.CopyFile : integer;
var
  ATime             : TFileTimestamps;
  ec                : integer;
  AAttr             : word;
begin
  ec:=OpenFiles (ATime,AAttr);
  if ec=errOK then begin
    DoProgress(acCopy,-sSource.Size,true);
    ec:=CopyStream(sSource,sDest,sSource.Size,true);
    if ec=errOk then ec:=CloseFiles(FCopyStorage,FCopyAltStreams,FCopyAcl,ATime,AAttr)
    else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// copy using shell function "IFileOperation" to include file infos
function TCopyThread.ShCopyFile : integer;
var
  FTime     : TFileTimestamps;
  NRead     : cardinal;
  quiet     : boolean;
  hr        : HResult;
begin
  Result:=errOK;
  FTime:=GetFileTimestamps(SourceName);
  if not FileExists(SourceName,false) then Result:=errError+errNotFound;
  if Result=errOK then begin
    if (length(DestName)=0) then Result:=errError+errFileCreate;
    end;
  if Result=errOK then begin
    quiet:=(scmQuiet in FShellCopy) or (LongFileSize(SourceName)<CMaxMemSize);
    DoProgress(acCopy,-1,false);   // Reset progress indicator
    if IsVista then hr:=IShellCopyFiles(Application.Handle,SourceName,DestName,quiet)
    else hr:=ShellCopyFiles(Application.Handle,SourceName,DestName,'',quiet);
    if CheckSysError(hr) then Result:=errError+errFileWrite;
    end;
  if FVerify and (Result=errOK) then begin   //Checksumme ermitteln
    UTotal:=0;
    try
      sSource:=TFileStream.Create(FilenameToXL(SourceName),fmOpenRead+fmShareDenyNone);
    except
      on E:EFOpenError do begin
        FSysErrMsg:=E.Message;
        Result:=errError+errFileVerify;
        end;
      end;
    if Result=errOK then begin
      DoProgress(acVerify,-sSource.Size,false);   // Reset progress indicator
      FCrc:=Crc32(0,nil,0);
      repeat
        try
          NRead:=sSource.Read(FBuffer[0],FBufSize);
          UTotal:=UTotal+NRead;
          DoProgress(acVerify,UTotal,true);
          FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        except
          on E:Exception do begin
            FSysErrMsg:=E.Message;
            Result:=errError+errFileVerify; NRead:=0;
            end;
          end;
        if FUserBreak then Result:=errUserBreak;
        until (NRead<FBufSize) or (Result<>errOK);
      FreeAndNil(sSource);
      end;
    end;
  if (Result=errOK) then begin
    if FCopyTimeStamp and FCopyCrTm then begin
      if CheckSysError(SetFileTimestamps(DestNameXL,FTime,FCheckTime,FCopyCrTm,false)) then
        Result:=errWarning+errFileTS;
      end;
    end;
  end;

function TCopyThread.GetZipSizes (fSource   : TStream;
                                  var CSize,USize : int64) : boolean;
var
  ZipInfo : TZipFileInfo;
begin
  Result:=GetZipLocalHeader(fSource,ZipInfo);
  if Result then begin
    CSize:=ZipInfo.CSize;
    USize:=ZipInfo.USize;
    end;
  fSource.Seek(0,soFromBeginning);
  end;

function TCopyThread.UnZipStream (fSource,fDest : TStream) : integer;
var
  mt : cardinal;
begin
  Result:=UnZipStream (fSource,fDest,mt);
  end;

function TCopyThread.UnZipStream (fSource,fDest : TStream; var ModTime : cardinal) : integer;
var
  ZipInfo   : TZipFileInfo;
  ec        : integer;
  sTmp      : TStream;
  tf        : boolean;
  tzn       : string;
begin
  ec:=errOK;
  if ec=errOK then begin
    if not GetZipLocalHeader (fSource,ZipInfo) then ec:=errFileGZip;
    if ec=errOK then begin
      ModTime:=FileAgeToUnixTime(ZipInfo.TimeStamp);
      if (ZipInfo.Flag and zipEncryptFlag)<>0 then begin  // decrypt
        if not assigned(FCrypting) then ec:=errError+errFileDCrypt
        else begin
          FCrypting.KeyMode:=ZipInfo.KeyMode;
          DoProgress(acDecrypt,-ZipInfo.CSize,FSpeedLimit);
          tf:=ZipInfo.CSize>CMaxMemSize;
          if tf then begin
            tzn:=FilenameToXL(fTmpDir+tmpEncName);
            sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
            end
          else sTmp:=TMemoryStream.Create;
          if FCrypting.DecryptStream(fSource,sTmp,ZipInfo.CSize,FSpeedLimit) then begin
            DoProgress(acUnCompress,-sTmp.Size,FSpeedLimit);
            sTmp.Seek(0,soFromBeginning);
            ec:=GzUncompress(sTmp,fDest,sTmp.size,FSpeedLimit);
            end
          else ec:=errError+errFileDCrypt;
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        end
      else begin
        DoProgress(acUnCompress,-ZipInfo.CSize,FSpeedLimit);
        if ZipInfo.Method=zipDeflate then ec:=GzUncompress(fSource,fDest,ZipInfo.CSize,FSpeedLimit)
        else ec:=CopyStream(fSource,fDest,ZipInfo.USize,FSpeedLimit);
        if (ec=errOK) and (FCrc<>ZipInfo.Crc) then ec:=errError+errFileRead;
        end;
      end;
    end;
  Result:=ec;
  end;

function TCopyThread.OpenFileForRead (const Filename : string) : integer;
var
  n,del : integer;
  se : string;
  ok : boolean;
begin
  n:=0; del:=1000*FMaxWait div 25;  // try 25 times until time out
  ok:=false;
  repeat
    try
      sDest:=TFileStream.Create(Filename,fmOpenRead+fmShareDenyNone);
      ok:=true;
    except
      on E:EFOpenError do se:=E.Message;
      end;
    if not ok then sleep(del);  // retry
    inc(n);
    until ok or (n>25);
  if ok then Result:=errOK
  else begin
    FSysErrMsg:=se+' ('+IntToStr(n*Del)+'ms)';
    Result:=errError+errVerOpen;
    end;
  end;

// verify check sum of copied file
function TCopyThread.VerifyFile : integer;
var
  ec        : integer;
  NRead,Crc : cardinal;
  RTotal    : int64;
begin
  RTotal:=0;
  ec:=OpenFileForRead(DestNameXL);
  if ec=errOK then begin
    DoProgress(acVerify,-UTotal,false);   // Reset progress indicator
    Crc:=Crc32(0,nil,0);
    repeat
      try
        NRead:=sDest.Read(FBuffer[0],FBufSize);
        RTotal:=RTotal+NRead;
        DoProgress(acVerify,RTotal,true);
        Crc:=ZipCrc32(Crc,FBuffer[0],NRead);
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          ec:=errError+errFileVerify; NRead:=0;
          end;
        end;
      if FUserBreak then ec:=errUserBreak;
      until (NRead<FBufSize) or (ec<>errOK);
    if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TCopyThread.Execute;
begin
  if scmCopy in FShellCopy then EndThread (ShCopyFile)
  else EndThread (CopyFile);
  end;

{ TEncryptThread ---------------------------------------------------------------- }
constructor TEncryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errEnCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TEncryptThread.Execute;
begin
  if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    EndThread (EncryptFile);
    end
  else EndThread (CopyFile);
  end;

{ TDecryptThread ---------------------------------------------------------------- }
constructor TDecryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errDeCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TDecryptThread.Execute;
begin
  if assigned(FCrypting) then EndThread (DecryptFile) else EndThread (CopyFile);
  end;

{ TGZipThread ------------------------------------------------------------------- }
constructor TGZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FName:=ExtractFilename(SourceName);
  FCompress:=true;
  FCrypting:=AEncryption;
  FErrFlag:=errGZip;
  end;

{ ------------------------------------------------------------------- }
// write gz header
// fEncMode = -1 - no encryption
//          =  0 - encryption with automatic keymode
//          >  0 - = keymode (1 = 128 bit, 3 = 256 bit)
function TGZipThread.GzWriteHeader (fDest : TStream;
                                    fName : string;
                                    fTime : cardinal;
                                    fEncMode : integer) : integer;
var
  Header : TGzHeader;
  n      : integer;
  w      : word;
  os     : byte;
  s      : RawByteString;
  ef     : TGzExtra64;
  cf     : TGzExtraCrypt;
begin
  Result:=errOK;       // o.k.
  if IsAnsiStr(fname) then begin    // gzip uses ISO-8859-1
    s:=fName; os:=OsDos;
    end
  else begin
    s:=UTF8Encode(fName); os:=OsNTFS;
    end;
  n:=length(s);
  with Header do begin
    Signature:=GzSignatur;
    Methode:=Z_DEFLATED;
    Flags:=gfExtraField;
    if n>0 then Flags:=Flags or gfFileName;
    if fEncMode>=0 then Flags:=Flags or gfEncrypted;
    TimeStamp:=fTime;
    XFlags:=0;
    OpSys:=os;
    end;
  try
    // write header
    if fDest.Write(Header,sizeof(TGzHeader))<sizeof(TGzHeader) then begin
      Result:=errError+errFileFull;  // dest disk full
      Exit;
      end;
    // write extra field with size information
    w:=sizeof(ef);
    if fEncMode>0 then inc(w,sizeof(cf));
    fDest.Write(w,2);
    with ef do begin
      SubfieldID:=ExtraGz64IDA; Size:=2*sizeof(int64);
      CSize:=0; USize:=0;
      end;
    if fDest.Write(ef,sizeof(ef))<sizeof(ef) then begin
      Result:=errError+errFileFull;  // dest disk full
      Exit;
      end;
    if fEncMode>=0 then begin
      with cf do begin
        SubfieldID:=ExtraGzCrypt; Size:=sizeof(KeyMode);
        KeyMode:=fEncMode;
        end;
      if fDest.Write(cf,sizeof(cf))<sizeof(cf) then begin
        Result:=errError+errFileFull;  // dest disk full
        Exit;
        end;
      end;
    if n>0 then begin   // write filename
      if fDest.Write(s[1],n+1)<n+1 then Result:=errError+errFileFull;  // dest disk full
      end;
  except
    on E:EWriteFileStreamError do begin
      FSysErrMsg:=E.Message;
      if E.ErrorCode=ERROR_DISK_FULL then Result:=errError+errFileFull // Ziel-Medium voll
      else Result:=errError+errFileWrite;
      end;
    end;
  end;

{ ------------------------------------------------------------------- }
// write gz trailer
function TGZipThread.GzWriteTrailer (fDest : TStream; ACrc : cardinal) : integer;
begin
  Result:=errOK;       // o.k.
  try
    with fDest do if Write(ACrc,4)<4 then Result:=errError+errFileFull  // dest disk full
    else if Write(TInt64(UTotal).Lo,4)<4 then Result:=errError+errFileFull  // dest disk full
    else begin  // write size info to extra field
      Position:=sizeof(TGzHeader)+6;
      Write(CTotal,sizeof(int64));
      Write(UTotal,sizeof(int64));
      end;
  except
    on E:Exception do begin
      FSysErrMsg:=E.Message;
      Result:=errError+errFileWrite; // error
      end;
    end;
  end;

{ ------------------------------------------------------------------- }
// compress file
function TGZipThread.CompressFile  : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  fp            : int64;
  Attr          : word;
  sTmp          : TStream;
  ec,km         : integer;
  FEnc,tf,sl    : boolean;
  tzn           : string;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    FEnc:=assigned(FCrypting);
    if FEnc then km:=FCrypting.KeyMode else km:=-1;
    // unix time for gzip
    utime:=FileTimeToUnixTime(FTime.LastWriteTime);
    ec:=GzWriteHeader(sDest,FName,utime,km);
    fp:=sDest.Position;
    if FEnc then begin
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=FilenameToXL(fTmpDir+tmpEncName);
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      sl:=false;
      end
    else begin
      sTmp:=sDest; sl:=true; tf:=false;
      end;
    // compress file
    if ec=errOK then begin
      DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
      ec:=GzCompress(sSource,sTmp,sl);
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then begin
          ec:=errError+errFileRead;
          FSysError:=GetLastError;
          end;
        // encrypt file
        if FEnc then begin
          if (ec=errOK) then begin
            DoProgress(acEncrypt,-sTmp.Size,false);   // Reset progress indicator
            sTmp.Seek(0,soFromBeginning);
            if (ec=errOK) and not FCrypting.EncryptStream(sTmp,sDest,true) then
              ec:=errError+errFileECrypt;
            end;
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        if ec=errOK then begin
          CTotal:=sDest.Position-fp;
//          if UTotal>MaxCardinal then UTotal:=MaxCardinal;
          if FEnc then ec:=GzWriteTrailer (sDest,0)
          else ec:=GzWriteTrailer (sDest,FCrc);
          end;
        end;
      end;
    if ec=errOK then ec:=CloseFiles(FCopyStorage,FCopyAltStreams,FCopyAcl,FTime,Attr)
    else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify check sum of copied file
function TGZipThread.VerifyGzFile : integer;
var
  fn            : string;
  ft,Crc        : cardinal;
  FEncMode      : integer;
  CSize,USize   : int64;
begin
  Result:=OpenFileForRead(DestNameXL);
  if Result=errOK then begin
    CrC:=FCrc;
    Result:=GetGzHeader (sDest,fn,ft,FEncMode,CSize,USize);
    DoProgress(acVerify,-USize,false);   // Reset progress indicator
    FCrc:=Crc32(0,nil,0); UTotal:=0;
    if Result=errOK then begin
      if FEncMode>=0 then begin // decrypt file
//        with sDest do n:=Size-Position-8;
        if not FCrypting.DecryptStream(sDest,nil,CSize,true) then Result:=errError+errFileVerify;
        end
      else begin
        Result:=GzUncompress(sDest,nil,true);
        if (Result=errOK) and (FCrc<>Crc) then Result:=errError+errFileVerify;
        end;
      end;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  end;

{ ------------------------------------------------------------------- }
// Copy, pack and encrypt from source stream to destination stream
function TGZipThread.ProcessStream (sSource : TStream; ACompress : boolean) : integer;
var
  fp     : int64;
  sTmp   : TStream;
  ec     : integer;
  tf     : boolean;
  tzn    : string;
begin
  fp:=sDest.Position;
  if assigned(FCrypting) then begin
    if (sSource.Size>0) and ACompress then begin
      DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=FilenameToXL(fTmpDir+tmpEncName);
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      ec:=GzCompress(sSource,sTmp,FSpeedLimit);  // size = 0, no compression
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then begin
          ec:=errError+errFileRead;
          FSysError:=GetLastError;
          end;
        end;
      // encrypt file
      DoProgress(acEncrypt,-sTmp.Size,false);      // Reset progress indicator
      if (ec=errOK) then begin
        sTmp.Seek(0,soFromBeginning);
        if not FCrypting.EncryptStream(sTmp,sDest,FSpeedLimit) then ec:=errError+errFileECrypt;
        end;
      sTmp.Free;
      if tf then DeleteFile(tzn);
      end
    else begin
      ec:=errOK; UTotal:=sSource.Size;
      // encrypt file
      DoProgress(acEncrypt,-sSource.Size,false);    // Reset progress indicator
      if not FCrypting.EncryptStream(sSource,sDest,FSpeedLimit) then ec:=errError+errFileECrypt;
      end;
    end
  else begin
    if (sSource.Size>0) and ACompress then begin
      DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
      ec:=GzCompress(sSource,sDest,FSpeedLimit)     // size = 0, no compression
      end
    else begin
      DoProgress(acCopy,-sSource.Size,false);       // Reset progress indicator
      ec:=CopyStream(sSource,sDest,sSource.Size,true);
      end;
    if ec=errOK then begin
      // z.B. wenn "sSource" gelockt ist (siehe LockFile)
      if UTotal<>sSource.Size then begin
        ec:=errError+errFileRead;
        FSysError:=GetLastError;
        end;
      end;
    end;
  CTotal:=sDest.Position-fp;
  Result:=ec;
  end;

function TGZipThread.CompressAsZip (ACompress : boolean) : integer;
var
  PkExtraCrypt : TPKExtraEncField;
  Attr      : word;
  ec        : integer;
  ZfName    : RawByteString;
  fn        : string;
  utime     : cardinal;
  LFHeader  : TPKLocalHeader;
  lef       : TPKExtraLocalField;
  FileData  : TWin32FindData;
  fp,ut     : int64;
  FEnc,
  z64,zae2  : boolean;
  rf,fc,el  : cardinal;
  pke       : TPkEndHeader;
  pkd       : TPkDirHeader;
  pkz       : TPKZip64CdRecord;
  pko       : TPKZip64CdLocator;
  ez64l,n   : integer;
begin
  PkExtraCrypt:=defPkExtraCrypt;
  FEnc:=assigned(FCrypting);
  if GetFileData(SourceNameXL,FileData) then begin
    utime:=GetFileAge(FileData);
    ut:=GetFileSize(FileData);
    Attr:=FileData.dwFileAttributes;
    z64:=ut>Z32MaxSize;
    zae2:=ut<AE1Size;
    with LFHeader do begin
      Signatur:=PKLocalSignatur;
      fn:=ExtractFilename(SourceName);
      if IsOEMStr(fn) then begin
        Flag:=0; ZfName:=fn;
        ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
        end
      else begin
        Flag:=zipUtf8Flag;
        ZfName:=UTF8Encode(fn);
        end;
      Flag:=Flag or zipPbFlag;  // set flag to indicate zip with single file
      if z64 then Vers:=VersZip64 else Vers:=VersDef;
      if FEnc then begin
        Flag:=Flag or zipEncryptFlag;
        Method:=zipEncrypted;
        ACompress:=true;         // compress always on encryption
        end
      else begin
        if (ut>0) and ACompress then begin
          Method:=zipDeflate;
          Flag:=Flag or zipMaxDeflFlag;
          end
        else Method:=zipNoComp;
        end;
      FTimeStamp:=utime;
      CRC:=0;
      FNLength:=length(ZfName);
      if z64 then begin
        with lef do begin
          Tag:=ExtraZip64;
          ESize:=16;
          Buffer[0]:=0;
          Buffer[1]:=0;
          end;
        CSize:=MaxCardinal;
        USize:=MaxCardinal;
        el:=20;
        ExtraLength:=20;
        end
      else begin
        CSize:=0;
        USize:=0;
        ExtraLength:=0;
        end;
      if FEnc then begin
        ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
        with PkExtraCrypt do begin
          if zae2 then Version:=VendorVersAE2 else Version:=VendorVersAE1;
          if (ut>0) and ACompress then Method:=zipDeflate else Method:=zipNoComp;
          Mode:=FCrypting.KeyMode;
          end;
        end;
      end;
    with sDest do try
      Write(LFHeader,sizeof(TPKLocalHeader));         // local header
      Write(ZfName[1],LFHeader.FNLength);             // filename
      if z64 then Write(lef,el);                      // extra Zip64 field
      if FEnc then Write(PkExtraCrypt,sizeof(TPKExtraEncField));  // extra enc field
      ec:=ProcessStream(sSource,ACompress);
      if ec=errOK then begin
        fc:=FCrc;
        if FEnc and zae2 then Position:=18    // position of compressed size
        else begin                            // Crc if no encryption or AE-1
          Position:=14;                       // position of crc
          Write(fc,sizeof(cardinal));
          end;
        if z64 then begin
          // write 8 byte values to extra field
          Position:=sizeof(TPKLocalHeader)+LFHeader.FNLength+4;
          Write (UTotal,sizeof(int64));
          Write (CTotal,sizeof(int64));
          end
        else begin
          // write 4 byte values
          rf:=TInt64(CTotal).Lo;
          Write(rf,sizeof(cardinal));
          rf:=TInt64(UTotal).Lo;
          Write(rf,sizeof(cardinal));
          end;
        Seek (0,soEnd);
        fp:=Position;
        // end directory
        with pkd do begin
          Signatur:=PkDirSignatur;
          Flag:=LFHeader.Flag;
          Method:=LFHeader.Method;
          FTimeStamp:=LFHeader.FTimeStamp;
          FNLength:=LFHeader.FNLength;
          if z64 then begin
            VersMade:=VersZip64+VersDos;
            VersExtr:=VersZip64+VersDos;
            with lef do begin
              Tag:=ExtraZip64;
              n:=0;
              if UTotal>MaxCardinal then begin
                USize:=MaxCardinal;
                Buffer[n]:=UTotal;
                inc(n);
                end
              else USize:=TInt64(UTotal).Lo;
              if CTotal>MaxCardinal then begin
                CSize:=MaxCardinal;
                Buffer[n]:=CTotal;
                inc(n);
                end
              else CSize:=TInt64(CTotal).Lo;
              Offset:=0;
              ESize:=8*n;
              ez64l:=4+ESize;
              ExtraLength:=ez64l;
              end;
            end
          else begin
            VersMade:=VersDef+VersDos;
            VersExtr:=VersDef;
            CSize:=TInt64(CTotal).Lo;
            USize:=TInt64(UTotal).Lo;
            Offset:=0;
            ExtraLength:=0;
            end;
          if FEnc then begin
            ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
            if PkExtraCrypt.Version=VendorVersAE2 then CRC:=0 else CRC:=FCRC;
            end
          else CRC:=FCRC;
          CommLength:=0;
          DiskStart:=0;
          IntAttr:=0;
          ExtAttr:=Attr;
          end;
        Write(pkd,sizeof(pkd));
        Write(ZfName[1],pkd.FNLength);
        if z64 then Write(lef,ez64l);  // extra fields
        if FEnc then Write(PkExtraCrypt,sizeof(TPKExtraEncField));
        end;
      with pke do begin
        Signatur:=PkEndSignatur;
        ThisDisk:=0;
        StartDisk:=0;
        ThisEntries:=1;
        TotalEntries:=1;
        CommLength:=0;
        end;  // check if file size > 4GB
      if (Position>MaxCardinal) then begin
        // Zip64 end of central directory record
        with pkz do begin
          Signatur:=Pk64RecSignatur;
          Size:=sizeof(TPKZip64CdRecord)-12;
          VersMade:=VersZip64+VersDos;
          VersExtr:=VersZip64+VersDos;
          ThisDisk:=0;
          StartDisk:=0;
          ThisEntries:=1;
          TotalEntries:=1;
          DirSize:=Position-fp;
          Offset:=fp;
          end;
        with pke do begin
          if pkz.DirSize>MaxCardinal then DirSize:=MaxCardinal
          else DirSize:=pkz.DirSize;
          Offset:=MaxCardinal;
          end;
        fp:=Position;
        // Zip64 end of central directory locator
        with pko do begin
          Signatur:=Pk64LocSignatur;
          StartDisk:=0;
          Offset:=fp;
          TotalDisk:=1;
          end;
        Write(pkz,sizeof(TPKZip64CdRecord));
        Write(pko,sizeof(TPKZip64CdLocator));
        end
      else with pke do begin
        DirSize:=Position-fp;
        Offset:=TInt64(fp).Lo;
        end;
      Write(pke,sizeof(pke));
    except
      on E:EWriteError do begin
        FSysErrMsg:=E.Message;
        ec:=errFileWrite;
        end;
      end;
    end
  else ec:=errFileRead;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
function TGZipThread.VerifyFromZip : integer;
var
  ZipInfo   : TZipFileInfo;
  ec        : integer;
begin
  ec:=OpenFileForRead(DestNameXL);
  if ec=errOK then begin
    // Header einlesen
    if not GetZipLocalHeader (sDest,ZipInfo) then ec:=errFileGZip
    else begin
      FCrc:=Crc32(0,nil,0); UTotal:=0;
      DoProgress(acVerify,-ZipInfo.USize,false);   // Reset progress indicator
      if assigned(FCrypting) then begin // decrypt file
        FCrypting.KeyMode:=ZipInfo.KeyMode;
        if not FCrypting.DecryptStream(sDest,nil,ZipInfo.CSize,FSpeedLimit) then ec:=errError+errFileVerify
        else ec:=errOK;
        end
      else with ZipInfo do begin
        if Method=zipDeflate then ec:=GzUncompress(sDest,nil,CSize,FSpeedLimit)
        else ec:=CopyStream(sDest,nil,CSize,FSpeedLimit);
        if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
        end;
      end;
    end;
  if assigned(sDest) then try FreeAndNil(sDest); except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TGZipThread.Execute;
begin
  EndThread (CompressFile);
  end;

{ TWinCopyThread ------------------------------------------------------------------- }
constructor TWinCopyThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                   AOptions : TCopyOptions; ASuspend : Boolean; ABufSize : integer;
                                   APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=coCompress in AOptions; FZipCompress:=coZipCompress in AOptions;
  FOverwrite:=coOverwrite in AOptions; FCopyStorage:=coStorage in AOptions;
  FCopyAltStreams:=coAltStreams in AOptions;
  FCopyAttr:=coAttributes in AOptions; FCopyTimeStamp:=coTimeStamp in AOptions;;
  FCopyAcl:=coPermissions in AOptions; FCopyCrTm:=coCreationTime in AOptions;
  FVerify:=coVerify in AOptions;
  FCheckTime:=coCheckTime in AOptions;
  NoWriteCache:=coNoCache in AOptions;
  if FZipCompress then FErrFlag:=errZip
  else if FCompress then FErrFlag:=errGZip
  else if assigned(FCrypting) then FErrFlag:=errEncrypt
  else FErrFlag:=errCopy;
  end;

procedure TWinCopyThread.Execute;
var
  ec : integer;
  FTime     : TFileTimestamps;
  Attr      : word;
begin
  if FVerify then SetNoWriteCache(true);
  if scmCopy in FShellCopy then begin
    ec:=ShCopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end
  else if FZipCompress then begin
    ec:=OpenFiles (FTime,Attr);
    if ec=errOK then begin
      ec:=CompressAsZip(FCompress);
      if ec=errOK then ec:=CloseFiles(false,FCopyAltStreams,FCopyAcl,FTime,Attr)
      else ReleaseStreams;
      if (ec=errOK) and FVerify then ec:=VerifyFromZip;
      end
    else ReleaseStreams;
    end
  else if FCompress then begin
    ec:=CompressFile;
    if FVerify and (ec=errOK) then ec:=VerifyGzFile;
    end
  else if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    ec:=EncryptFile;
    if FVerify and (ec=errOK) then ec:=VerifyEncFile;
    end
  else begin
    ec:=CopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end;
  EndThread (ec);
  end;

{ TGUnZipThread ------------------------------------------------------------------- }
constructor TGUnZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                  ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize,APriority);
  FCrypting:=AEncryption;
  FErrFlag:=errGUnzip;
  end;

{ ------------------------------------------------------------------- }
// read gz header
function TGUnZipThread.GzReadHeader (fSource   : TStream;
                                     var fName : string;
                                     var fTime : cardinal;
                                     var fEncMode    : integer;
                                     var CSize,USize : int64) : integer;
begin
  Result:=GetGzHeader (fSource,fName,fTime,fEncMode,CSize,USize);
  end;

function TGUnZipThread.GzSkipHeader (fSource : TStream) : integer;
var
  s : string;
  t : cardinal;
  e : integer;
  c,u : int64;
begin
  Result:=GetGzHeader(fSource,s,t,e,c,u);
  end;

function TGUnZipThread.GzReadSizes (fSource   : TStream;
                                    var fEncMode    : integer;
                                    var CSize,USize : int64) : integer;
var
  s : string;
  t : cardinal;
begin
  Result:=GetGzHeader(fSource,s,t,fEncMode,CSize,USize);
  end;

{ ------------------------------------------------------------------- }
// read gz trailer
function TGUnZipThread.GzReadTrailer (fSource : TStream; USize : int64) : integer;
var
  aCrc,FSize : cardinal;
begin
  Result:=errOK;       // o.k.
  try
    fSource.Read(aCrc,4);
    fSource.Read(FSize,4);
    if (FCrc<>aCrc) or (TInt64(USize).Lo<>FSize) or (USize<>UTotal) then begin
      Result:=errError+errFileCheck; exit; // checksum error
      end;
  except
    on E:Exception do begin
      FSysErrMsg:=E.Message;
      Result:=errError+errFileRead; // error
      end;
    end;
  end;

{ ------------------------------------------------------------------- }
// uncompress file
function TGUnZipThread.UncompressFile : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  ec            : integer;
  FName         : string;
  fp,n,sz       : int64;
  Attr          : word;
  sTmp          : TStream;
  FEncMode      : integer;
  tf            : boolean;
  tzn           : string;
begin
  ec:=errOK;
//  if length(SourceName)>MAX_PATH-1 then ec:=errError+errLongPath else
  if not FileExists(SourceNameXL,false) then ec:=errError+errNotFound;
  if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
  if (ec=errOK) then begin
    FTime:=GetFileTimestamps(SourceNameXL,false);
    Attr:=FileGetAttr(SourceNameXL,false);
    end;
  if ec=errOK then begin
    try
      sSource:=TFileStream.Create(SourceNameXL,fmOpenRead+fmShareDenyNone);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileOpen;
        end;
      end;
    end;
  if ec=errOK then begin
    ec:=GzReadHeader(sSource,FName,utime,FEncMode,n,sz);
    if (ec=errOK) then begin
//      if length(FName)>0 then DestName:=DestName+FName
//      else DestName:=DestName+ChangeFileExt(ExtractFilename(SourceName),'');
      // always overwrite destination, reset attributes
      if FOverwrite and FileExists(DestNameXL,false) then begin
        if FileSetAttr(DestNameXL,faArchive,false)<>0 then ec:=errError+errFileAttr;
        end;
      end;
    if ec=errOK then begin
      try
        sDest:=TWriteFileStream.Create(DestNameXL,FOverwrite,FWriteMode);
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          ec:=errError+errFileCreate;
          end;
        end;
      end;
    if (ec=errOK) and (FEncMode>=0) and not assigned(FCrypting) then ec:=errError+errFileDCrypt;
    if ec=errOK then begin
      fp:=sSource.Position;
      if FEncMode>=0 then begin // decrypt file
        DoProgress(acDecrypt,-sSource.Size,true);
        FCrypting.KeyMode:=FEncMode;
        tf:=sSource.Size>CMaxMemSize;
        if tf then begin
          tzn:=FilenameToXL(fTmpDir+tmpEncName);
          sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
          end
        else sTmp:=TMemoryStream.Create;
        with sSource do n:=Size-Position-8;
        if FCrypting.DecryptStream(sSource,sTmp,n,true) then begin
          DoProgress(acUnCompress,-sz,false);
          sTmp.Seek(0,soFromBeginning);
          ec:=GzUncompress(sTmp,sDest,sTmp.size,false);
          end
        else ec:=errError+errFileDCrypt;
        sTmp.Free;
        if tf then DeleteFile(tzn);
        // Ignore trailer
        end
      else begin
        DoProgress(acUnCompress,-sz,true);
        ec:=GzUncompress(sSource,sDest,true);
        if (ec=errOK) then begin
          CTotal:=sSource.Position-fp;
          ec:=GzReadTrailer (sSource,sz);
          end;
        end;
      end;
    if ec=errOK then begin
      if utime>0 then FTime.LastWriteTime:=UnixTimeToFileTime(utime);
      ec:=CloseFiles (false,false,FCopyAcl,FTime,Attr)
      end
    else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

procedure TGUnZipThread.Execute;
begin
  EndThread (UncompressFile);
  end;

{ ------------------------------------------------------------------- }
{ TUnZipThread ------------------------------------------------------------------- }
// unpack zip file with single compressed file
// ASourceName  = gz-file,  ADestFolder = destination folder
// ADecryption = Decrypting object (see AesLib), set to nil for plain uncompressing
constructor TUnZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                    ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize,APriority);
  FCrypting:=AEncryption;
  FErrFlag:=errGUnzip;
  end;

function TUnZipThread.UnZipFile : integer;
var
  FTime     : TFileTimestamps;
  Attr      : word;
begin
  Result:=errOK;
  if not FileExists(SourceNameXL,false) then Result:=errError+errNotFound;
  if (Result=errOK) and (length(DestName)=0) then Result:=errError+errFileCreate;
  if (Result=errOK) then begin
    FTime:=GetFileTimestamps(SourceNameXL,false);
    Attr:=FileGetAttr(SourceNameXL,false);
    end;
  if Result=errOK then begin
    try
      sSource:=TFileStream.Create(SourceNameXL,fmOpenRead+fmShareDenyNone);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        Result:=errError+errFileOpen;
        end;
      end;
    end;
  if Result=errOK then begin
    // always overwrite destination, reset attributes
    if FOverwrite and FileExists(DestNameXL,false) then begin
      if FileSetAttr(DestNameXL,faArchive,false)<>0 then Result:=errError+errFileAttr;
      end;
    end;
  if Result=errOK then begin
    try
      sDest:=TWriteFileStream.Create(DestNameXL,FOverwrite,FWriteMode);
    except
      on E:Exception do begin
        FSysErrMsg:=E.Message;
        Result:=errError+errFileCreate;
        end;
      end;
    end;
  if Result=errOK then Result:=UnZipStream(sSource,sDest);
  if Result=errOK then Result:=CloseFiles (false,false,FCopyAcl,FTime,Attr) else ReleaseStreams;
  end;

procedure TUnZipThread.Execute;
begin
  EndThread (UnZipFile);
  end;

{ ------------------------------------------------------------------- }
// Verify two files by comparing bytewise
// AVerifyName = file to be verified, may be compressed or encrypted
// ARefName = original file at source as reference
constructor TCompareThread.Create (const AVerifyName,ARefName : string;  AEncryption : TEncryption;
                                  ACompress,AZipCompress,ASuspend : Boolean; ABufSize : integer;
                                  APriority : TThreadPriority);
begin
  inherited Create (AVerifyName,ARefName,AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress; FZipCompress:=AZipCompress;
  if FZipCompress then FErrFlag:=errZip
  else if FCompress then FErrFlag:=errGUnZip
  else if assigned(FCrypting) then FErrFlag:=errDecrypt
  else FErrFlag:=0;
  end;

function TCompareThread.CompareDecryptedFile (fSource,fDest : TStream) : integer;
var
  tf          : boolean;
  ten         : string;
  sTmp        : TStream;
begin
  Result:=errOK;
  tf:=fSource.Size>UMaxMemSize;
  if tf then begin
    ten:=fTmpDir+tmpEncName;
    sTmp:=TFileStream.Create(FilenameToXL(ten),fmCreate); // write to disk
    end
  else sTmp:=TMemoryStream.Create;
  DoProgress(acDecrypt,-fDest.Size,true);
  Result:=ReadCryptHeader(fDest);
  if Result=errOK then begin
    if not FCrypting.DecryptStream(fDest,sTmp,fDest.Size-fDest.Position,true) then
      Result:=errError+errFileDCrypt;
    end;
  if Result=errOK then Result:=CompareStreams(fSource,sTmp,false);
  sTmp.Free;
  if tf then DeleteFile(ten);
  end;

function TCompareThread.CompareCompressedFile(fSource,fDest : TStream) : integer;
var
  tf          : boolean;
  tzn,ten     : string;
  sETmp,sZTmp : TStream;
  FEncMode    : integer;
  n,sz        : int64;
begin
  Result:=GzReadSizes(fDest,FEncMode,n,sz);
  if Result=errOK then begin
    tf:=fSource.Size>UMaxMemSize;
    if tf then begin
      tzn:=FilenameToXL(fTmpDir+tmpGzipName);
      sZTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
      end
    else sZTmp:=TMemoryStream.Create;
    if assigned(FCrypting) then begin
      DoProgress(acDecrypt,-sDest.Size,true);
      FCrypting.KeyMode:=FEncMode;
      if tf then begin
        ten:=fTmpDir+tmpEncName;
        sETmp:=TFileStream.Create(FilenameToXL(ten),fmCreate); // write to disk
        end
      else sETmp:=TMemoryStream.Create;
      with fDest do n:=Size-Position-8;
      if FCrypting.DecryptStream(fDest,sETmp,n,true) then begin
        DoProgress(acUnCompress,-sz,false);
        sETmp.Seek(0,soFromBeginning);
        Result:=GzUncompress(sETmp,sZTmp,sETmp.size,false);
        end
      else Result:=errError+errFileDCrypt;
      sETmp.Free;
      if tf then DeleteFile(ten);
      end
    else begin
      DoProgress(acUnCompress,-sz,true);
      Result:=GzUncompress(fDest,sZTmp,true);
      end;
    if Result=errOK then Result:=CompareStreams(fSource,sZTmp,false);
    sZTmp.Free;
    if tf then DeleteFile(tzn);
    end;
  end;

function TCompareThread.CompareZipFile(fSource,fDest : TStream) : integer;
var
  tf          : boolean;
  tzn         : string;
  sZTmp       : TStream;
begin
  tf:=fSource.Size>UMaxMemSize;
  if tf then begin
    tzn:=FilenameToXL(fTmpDir+tmpEncName);
    sZTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
    end
  else sZTmp:=TMemoryStream.Create;
  Result:=UnZipStream(fDest,sZTmp);
  if Result=errOK then Result:=CompareStreams(fSource,sZTmp,false);
  sZTmp.Free;
  if tf then DeleteFile(tzn);
  end;

procedure TCompareThread.Execute;
var
  ec : integer;
begin
  ec:=errOK;
  if not FileExists(DestNameXL,false) then ec:=errError+errNotFound;
  if (ec=errOK) and not FileExists(SourceNameXL,false) then ec:=errError2+errNotFound;
  if ec=errOK then begin
    try       // reference file
      sSource:=TFileStream.Create(DestNameXL,fmOpenRead+fmShareDenyNone);
    except
      on E:EFOpenError do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileOpen;
        end;
      end;
    end;
  if ec=errOK then begin
    try     // file to be verified
      sDest:=TFileStream.Create(SourceNameXL,fmOpenRead+fmShareDenyNone);
    except
      on E:EFOpenError do begin
        FSysErrMsg:=E.Message;
        ec:=errError2+errFileOpen;
        end;
      end;
    end;
  if ec=errOK then begin
    if FZipCompress then ec:=CompareZipFile(sSource,sDest)
    else if FCompress then ec:=CompareCompressedFile(sSource,sDest)
    else if assigned(FCrypting) then ec:=CompareDecryptedFile(sSource,sDest)
    else ec:=CompareStreams(sSource,sDest,true);
    end;
  ReleaseStreams;
  EndThread (ec);
  end;

{ ------------------------------------------------------------------- }
constructor TPkHeader.Create (const Fn : RawByteString; Ec,Al,Ts,Chk : cardinal;
                              AA,AF,AM,EV,ED,EM,SM : word; Ofs,CSz,USz : int64);
begin
  inherited Create;
  FName:=Fn;
  ErrCode:=Ec;
  Tag:=0;               // for internal use
  DirLength:=Al;        // total length of directory entry (only ReadDirectory)
  Timestamp:=Ts; CRC:=Chk; Attr:=AA; GpFlag:=AF; CompMethod:=AM;
  EncVers:=EV; EncMode:=ED; EncMeth:=EM; Segment:=SM;
  Offset:=Ofs; CSize:=CSz; USize:=USz;
  Zip64:=(Offset>MaxCardinal) or (CSize>MaxCardinal) or (USize>MaxCardinal);
  IsDir:=(CSz=0) and (USz=0); //(AF and not zipUtf8Flag=0);
  end;

// for rebuilding of Zip archives
constructor TPkHeader.Assign (APkHeader : TPkHeader; Ofs : int64);
begin
  inherited Create;
  FName:=APkHeader.FName;
  ErrCode:=APkHeader.ErrCode;
  Timestamp:=APkHeader.TimeStamp;
  CRC:=APkHeader.CRC;
  Attr:=APkHeader.Attr;
  GpFlag:=APkHeader.GpFlag;
  CompMethod:=APkHeader.CompMethod;
  EncVers:=APkHeader.EncVers;
  EncMode:=APkHeader.EncMode;
  EncMeth:=APkHeader.EncMeth;
  Segment:=APkHeader.Segment;
  Tag:=APkHeader.Tag;
  DirLength:=APkHeader.DirLength;
  Offset:=Ofs;                   // assign new position
  CSize:=APkHeader.CSize;
  USize:=APkHeader.USize;
  Zip64:=APkHeader.Zip64;
  IsDir:=APkHeader.IsDir;
  end;

{ ------------------------------------------------------------------- }
(* generate error message *)
constructor EZipError.Create (const ErrString : string; ErrCode : integer;
                              const ASysErrMsg : string);
begin
  inherited Create (ErrString);
  ZipError:=ErrCode or ErrZip;
  SysErrorMsg:=ASysErrMsg;
  end;

// =====================================================================
{ TZip  ----------------------------------------------------}
// Common Zip object
constructor TZip.Create (const ZipName : string; const Password : AnsiString;
                         AKeyMode,ABufSize,AMode : integer);
begin
  inherited Create;
  FStream:=TFileStream.Create(FilenameToXL(ZipName),AMode);
  FileList:=TStringList.Create;
  FileList.Sorted:=false;
  if ABufSize>0 then FBufferSize:=ABufSize
  else FBufferSize:=defBufferSize;
  FIsTemp:=false;     // if Zip file is temporary -> no speed limits
  FErrorFlag:=false;
  ZName:=ZipName;
  FBaseDir:='';
  PkExtraCrypt:=defPkExtraCrypt;
  FEncrypt:=false;
  InitEncryption(Password,AKeyMode);
  FNumSeg:=1; FCurSeg:=0;
  FPriority:=tpNormal;
  FOnProgress:=nil;
  FOnTimeout:=nil;
  fZipTmpDir:=TempDirectory;
  end;

destructor TZip.Destroy;
begin
  if FEncrypt then FEncryption.Free; // release encryption
  FileList.Free;
  try FStream.Free; except end;
  inherited Destroy;
  end;

procedure TZip.SetTempDirectory (const ATempDir : string);
begin
  if (length(ATempDir)>0) and DirectoryExists(ATempDir) then fZipTmpDir:=SetDirName(ATempDir)
  else fZipTmpDir:=TempDirectory;
  end;

procedure TZip.InitEncryption (const Password : AnsiString; AKeyMode : integer);
begin
  if FEncrypt then FEncryption.Free;
  FEncrypt:=length(Password)>0;
  if FEncrypt then begin   // initialize encryption
    FEncryption:=TEncryption.Create(Password,AKeyMode,FBufferSize);
    PkExtraCrypt.Mode:=FEncryption.KeyMode;
    end
  else FEncryption:=nil;
  end;

procedure TZip.SetProgress (ACallBack : TFileProgressEvent);
begin
  FOnProgress:=ACallBack;
  if FEncrypt then FEncryption.OnProgress:=ACallBack;
  end;

function TZip.GetEmpty : boolean;
begin
  Result:=FileList.Count=0;
  end;

function TZip.GetPosition : int64;
begin
  Result:=FStream.Position;
  end;

procedure TZip.WritePosition (Value : int64);
begin
  FStream.Position:=Value;
  end;

procedure TZip.DoProgress(AAction : TFileAction; ACount : int64; ASpeedLimit : boolean);
begin
  if assigned(FOnProgress) then begin
    FOnProgress(AAction,ACount,ASpeedLimit);
    Application.ProcessMessages;
    end;
  end;

function TZip.OpenNextSegment (var nSeg : word; var sSource : TFileStream) : integer;
var
  sn : string;
begin
  if nSeg<>fCurSeg then sSource.Free;
  Result:=errOK; inc(nSeg);
  if nSeg=fCurSeg then begin
    sSource:=FStream; sSource.Position:=0;
    end
  else begin
    sn:=GetSegmentName(ZName,nSeg);
    if FileExists(sn) then begin
      try
        sSource:=TFileStream.Create(sn,fmOpenRead);
      except
        Result:=errZipRdSeg;
        end;
      end
    else Result:=errZipRdSeg;
    end;
  end;

function TZip.OpenSegment (nSeg : word; var sSource : TFileStream): integer;
var
  sn : string;
begin
  Result:=errOK;
  if nSeg<>fCurSeg then begin
    if assigned(sSource) then sSource.Free;
    sn:=GetSegmentName(ZName,nSeg);
    if FileExists(sn) then begin
      try
        sSource:=TFileStream.Create(sn,fmOpenRead);
      except
        Result:=errZipRdSeg;
        end;
      end
    else Result:=errZipRdSeg;
    end;
  end;

function TZip.CopyStream (fSource,fDest : TStream; SLength : int64;
                          var nTotal : int64; var ErrMsg : string) : integer;
var
  cst : TCopyStreamThread;
begin
  cst:=TCopyStreamThread.Create(fSource,fDest,SLength,false,FBufferSize,FPriority);
  with cst do begin
    TempDir:=fZipTmpDir;
    OnProgress:=FOnProgress;
    Resume;
    repeat
      Application.ProcessMessages;
      Sleep(10);
      until Done;
    Result:=ErrorCode and errAllCodes;
    ErrMsg:=SysErrorMsg;
    end;
  cst.Free;
//var
//  FBuffer       : array of byte;
//  NRead,NWrite,
//  fb            : Integer;
//begin
//  SetLength(FBuffer,FBufferSize);
//  Result:=errOK; ErrMsg:=''; nTotal:=0;
//  repeat
//    if SLength<FBufferSize then fb:=SLength else fb:=FBufferSize;
//    try
//      NRead:=fSource.Read(FBuffer[0],fb);
//      if NRead<fb then begin
//        Result:=errFileRead;   // z.B. wenn "sSource" gelockt ist
//        ErrMsg:=GetLastErrMsg;
//        end;
//    except
//      on E:Exception do begin
//        ErrMsg:=E.Message;
//        Result:=errFileRead;
//        end;
//      end;
//    if Result=errOK then begin
//      inc(nTotal,NRead);
//      DoProgress(acCopy,nTotal,false);
//      try
//        NWrite:=fDest.Write(FBuffer[0],NRead);
//        if NWrite<NRead then Result:=errFileFull; // Ziel-Medium voll
//      except
//        on E:Exception do begin
//          ErrMsg:=E.Message;
//          Result:=errFileWrite;
//          end;
//        end;
//      end;
//    dec(Slength,NRead);
//    until (SLength<=0) or (Result<>errOK);
//  FBuffer:=nil;
  end;

procedure TZip.CopyToZip(AZip : TZip; SLength : int64);
var
  n : int64;
  s : string;
begin
  CopyStream(FStream,AZip.FStream,SLength,n,s);
  end;

{ ------------------------------------------------------------------- }
// Read Zip archive
constructor TZipRead.Create (const ZipName : string; const Password : AnsiString;
                             AKeyMode,ABufSize : integer);
begin
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmOpenRead);
  except
    on E:EFCreateError do begin
      raise EZipError.Create (Format(rsErrOpening,[ZipName]),errError+errFileOpen,E.Message);
      end;
    end;
  end;

destructor TZipRead.Destroy;
var
  i : integer;
begin
  with FileList do for i:=0 to Count-1 do
    if assigned(Objects[i]) then Objects[i].Free;
  inherited Destroy;
  end;

// Read central directory
procedure TZipRead.ReadDirectory;
var
  pke   : TPkEndHeader;
  pkd   : TPkDirHeader;
  pkz   : TPKZip64CdRecord;
  pko   : TPKZip64CdLocator;
  ZfName     : RawByteString;
  s,sp       : string;
  usz,csz,fp : int64;
  i,ne,ec,
  te,el,k      : cardinal;
  pkelf        : TExtraFieldInfo;
  Buffer       : array of byte;
  fd,
  EV,ED,EM,SM  : word;
  sSeg         : TFileStream;
begin
  FStream.Position:=FStream.Size-sizeof(pke);
  FStream.Read(pke,sizeof(pke));   // end header
  with pke do begin
    if Signatur<>PkEndSignatur then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
      Exit;
      end;
    if (pke.Offset=MaxCardinal) or (pke.DirSize=MaxCardinal) or (pke.TotalEntries=MaxWord) then begin // Zip64 extension
      FStream.Position:=FStream.Size-sizeof(pke)-sizeof(pko);  // read Zip64 end of central directory locator
      FStream.Read(pko,sizeof(pko));
      if pko.Signatur<>Pk64LocSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        exit;
        end;
      FStream.Position:=pko.Offset;                    // read Zip64 end of central directory record
      FStream.Read(pkz,sizeof(pkz));
      if pkz.Signatur<>Pk64RecSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Exit;
        end;
      with pkz do begin
        fp:=Offset;
        ne:=TotalEntries;
        FNumSeg:=ThisDisk;
        fd:=StartDisk;       // start segment of central directory
        te:=ThisEntries;
        end;
      end
    else with pke do begin
      fp:=Offset;
      ne:=TotalEntries;
      FNumSeg:=ThisDisk;
      fd:=StartDisk;        // dart segment of central directory
      te:=ThisEntries;
      end;
    end;
  ec:=errOK; fCurSeg:=FNumSeg;
  if fd<>FNumSeg then ec:=OpenSegment(fd,sSeg)
  else sSeg:=FStream;
  sSeg.Position:=fp;
  bod:=fp;
  sp:='';
  if ec=errOK then for i:=0 to ne-1 do begin  // read directory entries
    with sSeg do if Position>=Size then ec:=OpenNextSegment(fd,sSeg);
    if ec<>errOK then Break;
    try
      sSeg.Read(pkd,sizeof(pkd));
    except
      on E:Exception do begin
        ec:=errFileRead;
        Break;
        end;
      end;
    with pkd do begin
      if Signatur<>PkDirSignatur then begin
        ec:=errSignature;
//        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Break;
        end;
//      if not ((VersExtr=VersDef) or (VersExtr=VersZip64)) then ec:=errFormat;
      if not ((Method=zipNoComp) or (Method=zipDeflate) or (Method=zipEncrypted)) then ec:=errFormat;
      if ec<>errOK then Break;
      fp:=Offset;
      usz:=USize;
      csz:=CSize;
      SM:=DiskStart;
      SetLength(ZfName,FNLength);
      try
        sSeg.Read(ZfName[1],FNLength);
      except
        on E:Exception do begin
          ec:=errFileRead;
          Break;
          end;
        end;
      if Flag and zipUtf8Flag<>0 then s:=UnixPathToDosPath(UTF8ToWideString(ZfName))
      else s:=UnixPathToDosPath(StrToAnsi(ZfName));
      if ExtraLength>0 then begin      // read extra fields
        el:=ExtraLength;
        try
          while el>0 do begin  // read extra fields
            sSeg.Read(pkelf,SizeOf(TExtraFieldInfo));
            with pkelf do begin
              SetLength(Buffer,ESize);
              sSeg.Read(Buffer[0],ESize);
              k:=0;
              if Tag=ExtraZip64 then begin
                if USz=MaxCardinal then begin
                  Move(Buffer[k],USz,8);
                  inc(k,8);
                  end;
                if (CSz=MaxCardinal) and (ESize>k) then begin
                  Move(Buffer[k],CSz,8);
                  inc(k,8);
                  end;
                if (fp=MaxCardinal) and (ESize>k) then begin
                  Move(Buffer[k],fp,8);
                  end;
                end
              else if Tag=ExtraCrypt then begin
                Move(Buffer[0],EV,2);  // Version
                ED:=Buffer[4];         // Mode
                Move(Buffer[5],EM,2);  // Methode
                end;
              el:=el-SizeOf(TExtraFieldInfo)-ESize;
              end;
            end;
        except
          on E:Exception do begin
            ec:=errFileRead;
            Break;
            end;
          end;
        end;
      FileList.AddObject(s,TPkHeader.Create(ZfName,ec,sizeof(pkd)+FNLength+ExtraLength,
                         FTimeStamp,CRC,ExtAttr,Flag,Method,EV,ED,EM,SM,fp,CSz,USz));
      end;
    end;
  if fd<>fCurSeg then FreeAndNil(sSeg);
  if ec<>errOK then raise EZipError.Create (Format(rsErrReading,[ZName]),errError+ec);
  end;

function TZipRead.Filename (Index : integer) :string;
begin
  with FileList do if (Index>=0) and (Index<Count) then Result:=Strings[Index]
  else Result:='';
  end;

function TZipRead.IsEncrypted(Index : integer) : boolean;
begin
  Result:=false;
  with FileList do if (Index>=0) and (Index<Count) then
    Result:=((Objects[Index] as TPkHeader).GpFlag and zipEncryptFlag)<>0;
  end;

function TZipRead.ReadLocalHeader (sSource : TStream; AOfs : int64; var ZipInfo : TZipFileInfo) : boolean;
begin
  sSource.Position:=AOfs;
  Result:=GetZipLocalHeader(sSource,ZipInfo);
  if assigned(FEncryption) then FEncryption.KeyMode:=ZipInfo.KeyMode;
  end;


function TZipRead.ProcessRead (sSource : TStream; sDest : TFileStream;
                               AZipInfo : TZipFileInfo; var ErrMsg : string) : integer;
var
  timo : boolean;
begin
  ZipThread:=TZipReadThread.Create(sSource,sDest,FEncryption,AZipInfo,false,not FIsTemp,
                        FBufferSize,FPriority,FOnProgress);
  with ZipThread do begin
    repeat
      Application.ProcessMessages;
      Sleep(10);
      if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
      until Done or timo;
    Result:=ErrorCode and errAllCodes;
    ErrMsg:=SysErrorMsg;
    Free;
    end;
  ZipThread:=nil;
  sDest.Free;
  end;

function TZipRead.ExtractSplittedFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
var
  tzn     : string;
  sTmp    : TStream;
  sSeg    : TFileStream;
  tf      : boolean;
  ZipInfo : TZipFileInfo;
  seg     : word;
  n,nt    : int64;
begin
  if PkHeader.Segment=fCurSeg then Result:=ExtractFile(PkHeader,sDest,ErrMsg)
  else begin
    seg:=PkHeader.Segment; sSeg:=nil;
    Result:=OpenSegment(seg,sSeg);
    if Result=errOK then begin
      if ReadLocalHeader(sSeg,PkHeader.Offset,ZipInfo) then begin
        if ZipInfo.CSize+sSeg.Position<sSeg.Size then Result:=ProcessRead(sSeg,sDest,ZipInfo,ErrMsg)
        else begin        // need more than one segment
          tf:=PkHeader.USize>UMaxMemSize;
          if tf then begin
            tzn:=FilenameToXL(fZipTmpDir+tmpZipName);
            sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
            end
          else sTmp:=TMemoryStream.Create;
          n:=sSeg.Size-sSeg.Position;
          Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
          while (Result=errOK) and (sTmp.Size<PkHeader.CSize) do begin  // copy to sTmp
            Result:=OpenNextSegment(seg,sSeg);
            n:=PkHeader.CSize-n;
            if n>sSeg.Size then n:=sSeg.Size;
            if Result=errOK then Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
            end;
          sTmp.Position:=0;
          if Result=errOK then Result:=ProcessRead(sTmp,sDest,ZipInfo,ErrMsg);
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        end
      else begin
        Result:=errFileRead; ErrMsg:=GetLastErrMsg;
        end;
      end;
    if seg<>fCurSeg then FreeAndNil(sSeg);
    end;
  end;

function TZipRead.ExtractFile (PkHeader : TPkHeader; sDest : TFileStream;
                               var ErrMsg : string) : integer;
var
  ZipInfo : TZipFileInfo;
begin
  if ReadLocalHeader(FStream,PkHeader.Offset,ZipInfo) then begin
    Result:=ProcessRead(FStream,sDest,ZipInfo,ErrMsg);
    end
  else begin
    Result:=errFileRead; ErrMsg:=GetLastErrMsg;
    end;
  end;

procedure TZipRead.ExtractToFile (pkHeader : TPkHeader; const DestFilename : string);
var
  ec : integer;
  sd,se   : string;
  sDest : TFileStream;
begin
  sd:=FilenameToXL(DestFilename); ec:=errOK;
  if FileExists(sd,false) then begin
    if FileSetAttr(sd,faArchive,false)<>0 then ec:=errError+errFileAttr;
    end;
  if ec=errOK then begin
    try
      sDest:=TWriteFileStream.Create(sd);
    except
      on E:Exception do begin
        se:=E.Message;
        ec:=errError+errFileCreate;
        end;
      end;
    end;
  if ec=errOK then begin
    if fNumSeg>0 then ec:=ExtractSplittedFile(pkHeader,sDest,se) // splitted archive
    else ec:=ExtractFile(pkHeader,sDest,se);
    end;
  if ec=errOK then begin
    if SetFileAgeXL(DestFileName,pkHeader.TimeStamp)=0 then begin
      if FileSetAttrXL(DestFileName,pkHeader.Attr,false)>0 then
        raise EZipError.Create (Format(rsErrSetAttr,[DestFileName]),errWarning+errFileAttr,GetLastErrMsg);
      end
    else raise EZipError.Create (Format(rsErrTimeStamp,[DestFileName]),errWarning+errFileTS,GetLastErrMsg);
    end
  else begin
  // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

procedure TZipRead.ExtractToFile(Index : integer; const DestFilename : string);
var
  pkh : TPkHeader;
begin
  if assigned(FileList) then if (Index>=0) and (Index<FileList.Count) then begin
    pkh:=(FileList.Objects[Index] as TPkHeader);
    if ((pkh.GpFlag and zipEncryptFlag)<>0) and not assigned(FEncryption) then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errFileDCrypt);
      Exit;
      end;
    ExtractToFile(pkh,DestFileName);
    end
  else begin
  // set error flag on write errors
    raise EZipError.Create (GetCopyErrMsg(errError+errFileRead),errError+errFileRead,'');
    end;
  end;

// Extract to stream (< 4GB, no encryption, no splitting)
function TZipRead.ExtractToStream(const Filename : string; sDest : TStream) : boolean;
var
  n,ec   : integer;
  timo   : boolean;
  se     : string;
  ZipInfo : TZipFileInfo;
begin
  if assigned(FileList) then n:=FileList.IndexOf(Filename) else n:=-1;
  Result:=n>=0;
  if Result then begin
    FStream.Position:=(FileList.Objects[n] as TPkHeader).Offset;
    ZipThread:=TZipReadThread.Create(FStream,sDest,false,not FIsTemp,
                          FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      ec:=ErrorCode and errAllCodes;
      se:=SysErrorMsg;
      Free;
      end;
    ZipThread:=nil;
    sDest.Position:=0;
    if ec<>errOK then raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

function TZipRead.ProcessCompare (sSource,sDest : TStream; AZipInfo : TZipFileInfo;
                                  var ErrMsg : string) : integer;
var
  timo : boolean;
begin
  ZipThread:=TZipCompareThread.Create(sSource,sDest,FEncryption,AZipInfo,false,not FIsTemp,
                           FBufferSize,FPriority,FOnProgress);
  with ZipThread do begin
    repeat
      Application.ProcessMessages;
      Sleep(10);
      if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
      until Done or timo;
    Result:=ErrorCode and errAllCodes;
    ErrMsg:=SysErrorMsg;
    Free;
    end;
  ZipThread:=nil;
  end;

function TZipRead.CompareSplittedFile (PkHeader : TPkHeader; sDest : TFileStream;
                          var ErrMsg : string) : integer;
var
  tzn     : string;
  sTmp    : TStream;
  sSeg    : TFileStream;
  tf      : boolean;
  ZipInfo : TZipFileInfo;
  seg     : word;
  n,nt    : int64;
begin
  if PkHeader.Segment=fCurSeg then Result:=CompareFile(PkHeader,sDest,ErrMsg)
  else begin
    seg:=PkHeader.Segment; sSeg:=nil;
    Result:=OpenSegment(seg,sSeg);
    if Result=errOK then begin
      if ReadLocalHeader(sSeg,PkHeader.Offset,ZipInfo) then begin
        if ZipInfo.CSize+sSeg.Position<sSeg.Size then Result:=ProcessCompare(sSeg,sDest,ZipInfo,ErrMsg)
        else begin        // need more than one segment
          tf:=PkHeader.USize>UMaxMemSize;
          if tf then begin
            tzn:=FilenameToXL(fZipTmpDir+tmpZipName);
            sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
            end
          else sTmp:=TMemoryStream.Create;
          n:=sSeg.Size-sSeg.Position;
          Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
          while (Result=errOK) and (sTmp.Size<PkHeader.CSize) do begin  // copy to sTmp
            Result:=OpenNextSegment(seg,sSeg);
            n:=PkHeader.CSize-n;
            if n>sSeg.Size then n:=sSeg.Size;
            if Result=errOK then Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
            end;
          sTmp.Position:=0;
          if Result=errOK then Result:=ProcessCompare(sTmp,sDest,ZipInfo,ErrMsg);
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        end
      else begin
        Result:=errFileRead; ErrMsg:=GetLastErrMsg;
        end;
      end;
    if seg<>fCurSeg then FreeAndNil(sSeg);
    end;
  end;

function TZipRead.CompareFile (PkHeader : TPkHeader; sDest : TFileStream;
                               var ErrMsg : string) : integer;
var
  ZipInfo : TZipFileInfo;
begin
  if ReadLocalHeader(FStream,PkHeader.Offset,ZipInfo) then begin
    Result:=ProcessCompare(FStream,sDest,ZipInfo,ErrMsg);
    end
  else begin
    Result:=errFileRead; ErrMsg:=GetLastErrMsg;
    end;
  end;

procedure TZipRead.CompareToFile (pkHeader : TPkHeader; const RefFilename : string);
var
  ec    : integer;
  se,sd : string;
  sDest : TFileStream;
begin
  sd:=FilenameToXL(RefFilename); ec:=errOK;
  if not FileExists(sd,false) then ec:=errError+errNotFound
  else begin
    try
      sDest:=TFileStream.Create(sd,fmOpenRead+fmShareDenyNone);
    except
      on E:Exception do begin
        se:=E.Message;
        ec:=errError+errFileCreate;
        end;
      end;
    end;
  if ec=errOK then begin
    if fNumSeg>0 then ec:=CompareSplittedFile(pkHeader,sDest,se) // splitted archive
    else ec:=CompareFile(pkHeader,sDest,se);
    end;
  if ec<>errOK then begin
  // set error flag on compare errors
    FErrorFlag:=ec=errFileVerify;
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

procedure TZipRead.CompareToFile(Index : integer; const DestFilename : string);
var
  pkh : TPkHeader;
begin
  if assigned(FileList) then if (Index>=0) and (Index<FileList.Count) then begin
    pkh:=(FileList.Objects[Index] as TPkHeader);
    if ((pkh.GpFlag and zipEncryptFlag)<>0) and not assigned(FEncryption) then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errFileDCrypt);
      Exit;
      end;
    CompareToFile(pkh,DestFileName);
    end;
  end;

procedure TZipRead.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

{ ------------------------------------------------------------------- }
// Create ZIP archive
constructor TZipWrite.Create (const ZipName,BasicDirectory : string;
                              AMaxSize : int64;   // split zip file
                              const Password : AnsiString;
                              AKeyMode,ABufSize : integer);
var
  sg : cardinal;
begin
  fMaxSize:=AMaxSize;
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmCreate);
    FBaseDir:=IncludeTrailingPathDelimiter(BasicDirectory);
  except
    on E:EFCreateError do begin
      raise EZipError.Create(Format(rsErrCreating,[ZipName]),errError+errFileCreate,E.Message);
      end;
    end;
  FOnVerifyProgress:=nil; FOnFullDisk:=nil;
  if fMaxSize>0 then begin
    sg:=PkSplitSignatur;
    FStream.Write(sg,4);
    end;
  end;

constructor TZipWrite.Create (const ZipName,BasicDirectory : string;
                              const Password : AnsiString;
                              AKeyMode,ABufSize : integer);
begin
  Create(ZipName,BasicDirectory,0,Password,AKeyMode,ABufSize);
  end;

//  Write directory and final block of ZIP archive
destructor TZipWrite.Destroy;
var
  i : integer;
begin
  if assigned(FileList) then with FileList do for i:=0 to Count-1 do Objects[i].Free;
  inherited Destroy;
  end;

procedure TZipWrite.DeleteZipFile;
var
  i : integer;
begin
  try FStream.Free; except end;
  if fCurSeg>0 then begin // splitted
    for i:=0 to fCurSeg-1 do DeleteExistingFile(GetSegmentName(ZName,i));
    end;
  DeleteFileXL(ZName);
  end;

procedure TZipWrite.RenameZipFile (const NewName : string);
var
  i : integer;
begin
  try FStream.Free; except end;
  if fCurSeg>0 then begin // splitted
    for i:=0 to fCurSeg-1 do
      RenameFile(GetSegmentName(ZName,i),FilenameToXL(NewExt(NewName,'z'+Format('%.2d',[i+1]))));
    end;
  RenameFileXL(ZName,NewName);
  end;

function TZipWrite.NewSegment : integer;
var
  sn : string;
begin
  Result:=errOK;
  if fMaxSize>0 then begin
    FStream.Free;
    sn:=ExtractFileDrive(ZName);
    if (length(sn)>0) and (sn[1]<>'\') then begin
      if (DiskFree(byte(sn[1]) and $DF-64)<=fMaxSize) and assigned(FOnFullDisk) then begin
        if FOnFullDisk then Result:=errFileFull;  // cancel
        end;
      end;
    if Result=errOK then begin
      sn:=GetSegmentName(ZName,FCurSeg); //FilenameToXL(NewExt(ZName,'z'+Format('%.2d',[FCurSeg+1])));
      if DeleteExistingFile(sn) and RenameFile(FilenameToXL(ZName),sn) then begin
        try
          FStream:=TFileStream.Create(FilenameToXL(ZName),fmCreate);
        except
          Result:=errZipCrSeg;
          end;
        end
      else Result:=errZipCrSeg;
      if Result=errOK then inc(FCurSeg);
      end;
    end;
  end;

function TZipWrite.CheckSegment (ASpace: Longint): boolean;
begin
  if fMaxSize>0 then Result:=fMaxSize<FStream.Position+ASpace
  else Result:=false;
  end;

procedure TZipWrite.WriteDirectory;
var
  pke : TPkEndHeader;
  pkd : TPkDirHeader;
  pkl : TPkHeader;
  pkz : TPKZip64CdRecord;
  pko : TPKZip64CdLocator;
  sz,
  fp  : int64;
  ez64l,
  td,nd,
  i,n : integer;
  lef : TPKExtraLocalField;
begin
  fp:=FStream.Size; td:=fCurSeg; nd:=0; sz:=0;
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    pkl:=Objects[i] as TPkHeader;
    with pkd do begin
      Signatur:=PkDirSignatur;
      Flag:=pkl.GpFlag;
      Method:=pkl.CompMethod;
      FTimeStamp:=pkl.TimeStamp;
      FNLength:=length(pkl.FName);
      if pkl.Zip64 then begin
        VersMade:=VersZip64+VersDos;
        VersExtr:=VersZip64+VersDos;
        with lef do begin
          Tag:=ExtraZip64;
          n:=0;
          if pkl.USize>MaxCardinal then begin
            USize:=MaxCardinal;
            Buffer[n]:=pkl.USize;
            inc(n);
            end
          else USize:=TInt64(pkl.USize).Lo;
          if pkl.CSize>MaxCardinal then begin
            CSize:=MaxCardinal;
            Buffer[n]:=pkl.CSize;
            inc(n);
            end
          else CSize:=TInt64(pkl.CSize).Lo;
          if pkl.Offset>MaxCardinal then begin
            Offset:=MaxCardinal;
            Buffer[n]:=pkl.Offset;
            inc(n);
            end
          else Offset:=TInt64(pkl.Offset).Lo;
          ESize:=8*n;
          ez64l:=4+ESize;
          ExtraLength:=ez64l;
          end;
        end
      else begin
        VersMade:=VersDef+VersDos;
        VersExtr:=VersDef;
        CSize:=TInt64(pkl.CSize).Lo;
        USize:=TInt64(pkl.USize).Lo;
        Offset:=TInt64(pkl.Offset).Lo;
        ExtraLength:=0;
        end;
      if FEncrypt then begin
        ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
        with PkExtraCrypt do begin
          Version:=pkl.EncVers;
          if Version=VendorVersAE2 then CRC:=0 else CRC:=pkl.CRC;
          Method:=pkl.EncMeth;
          end;
        end
      else CRC:=pkl.CRC;
      CommLength:=0;
      DiskStart:=pkl.Segment;
      IntAttr:=0;
      ExtAttr:=pkl.Attr;
      end;
    try
      // check for space in segment
      n:=sizeof(pkd)+pkd.FNLength+pkd.ExtraLength;
      if FEncrypt then inc(n,sizeof(TPKExtraEncField));
      inc(sz,n); // size of central directory
      if CheckSegment(n) then begin
        if NewSegment<>errOK then raise Exception(GetLastErrMsg);
        nd:=i;
        end;
      FStream.Write(pkd,sizeof(pkd));
      FStream.Write(pkl.FName[1],pkd.FNLength);
      if pkl.Zip64 then FStream.Write(lef,ez64l);  // extra fields
      if FEncrypt then FStream.Write(PkExtraCrypt,sizeof(TPKExtraEncField));
    except
      on E:Exception do begin
        raise EZipError.Create (Format(rsErrDirEntry,[Strings[i]]),errError+errFileWrite,E.Message);
        exit;
        end;
      end;
    end;
  with pke do begin
    Signatur:=PkEndSignatur;
    ThisDisk:=fCurSeg;
    StartDisk:=td;
    if FileList.Count>MaxWord then begin
      ThisEntries:=MaxWord;
      TotalEntries:=MaxWord;
      end
    else begin
      ThisEntries:=FileList.Count-nd;
      TotalEntries:=FileList.Count;
      end;
    CommLength:=0;
    end;  // check if file size > 4GB
  if (FStream.Position>MaxCardinal) or (FileList.Count>MaxWord) or (sz>MaxCardinal) then begin
    // Zip64 end of central directory record
    with pkz do begin
      Signatur:=Pk64RecSignatur;
      Size:=sizeof(TPKZip64CdRecord)-12;
      VersMade:=VersZip64+VersDos;
      VersExtr:=VersZip64+VersDos;
      ThisDisk:=fCurSeg;
      StartDisk:=td;
      ThisEntries:=FileList.Count-nd;
      TotalEntries:=FileList.Count;
      DirSize:=sz;
      Offset:=fp;
      end;
    with pke do begin
      ThisDisk:=$FFFF;
      StartDisk:=$FFFF;
      if pkz.DirSize>MaxCardinal then DirSize:=MaxCardinal
      else DirSize:=pkz.DirSize;
      Offset:=MaxCardinal;
      end;
    fp:=FStream.Position;
    // Zip64 end of central directory locator
    with pko do begin
      Signatur:=Pk64LocSignatur;
      StartDisk:=0;
      Offset:=fp;
      TotalDisk:=1;
      end;
    try
      FStream.Write(pkz,sizeof(TPKZip64CdRecord));
      FStream.Write(pko,sizeof(TPKZip64CdLocator));
    except
      on E:Exception do begin
        raise EZipError.Create (rsErrEndRecord,errError+errFileWrite,E.Message);
        exit;
        end;
      end;
    end
  else with pke do begin
    DirSize:=sz;
    Offset:=TInt64(fp).Lo;
    end;
  try
    FStream.Write(pke,sizeof(pke));
  except
    on E:Exception do
      raise EZipError.Create (rsErrEndRecord,errError+errFileWrite,E.Message);
    end;
  end;

procedure TZipWrite.Verify;  // Verify zip archive
var
  i,ec : integer;
  se : string;
begin
  if assigned(FOnVerifyProgress) then FOnVerifyProgress(acZipVerify,-FileList.Count,not FIsTemp);
  FOnProgress:=nil;
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    if assigned(FOnVerifyProgress) then FOnVerifyProgress(acZipVerify,i,not FIsTemp);
    if fMaxSize>0 then begin   // splitted
      ec:=VerifySplittedFile(ZName,Objects[i] as TPkHeader,se);
      end
    else ec:=VerifyFile(Objects[i] as TPkHeader,se);
    if ec<>errOK then begin
      raise EZipError.Create (Format(rsErrVerify,[Strings[i]]),ec,se);
      Break;
      end;
    end;
  end;

//  Add directory name to archive
procedure TZipWrite.AddDir (DirName : string);
var
  ZfName     : RawByteString;
  fn,se      : string;
  ec         : integer;
  LFHeader   : TPKLocalHeader;
  fp         : int64;
begin
  if DirectoryExistsXL(DirName,false) then begin
    fn:=ExtractRelativePath(FBaseDir,DirName);
    se:='';
    with LFHeader do begin
      Signatur:=PKLocalSignatur;
      Vers:=VersDef;
      fn:=ExtractRelativePath(FBaseDir,DirName);
      if IsOEMStr(fn) then begin
        Flag:=0; ZfName:=fn;
        ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
        end
      else begin
        Flag:=zipUtf8Flag;
        ZfName:=UTF8Encode(fn);
        end;
      Method:=zipNoComp;
      FTimeStamp:=DirectoryAge(ExcludeTrailingPathDelimiter(DirName));
      CRC:=0;
      FNLength:=length(ZfName);
      CSize:=0;
      USize:=0;
      ExtraLength:=0;
      end;
    fp:=FStream.Position;
    try
      // check for space in segment
      if CheckSegment(sizeof(TPKLocalHeader)) then ec:=NewSegment else ec:=errOK;;
      if ec=errOK then begin
        FStream.Write(LFHeader,sizeof(TPKLocalHeader));       // local header
        FStream.Write(ZfName[1],LFHeader.FNLength);             // filename
        FileList.AddObject(fn,TPkHeader.Create(ZfName,0,0,LfHeader.FTimeStamp,0,
                 FileGetAttr(DirName),LfHeader.Flag,LfHeader.Method,0,0,0,FCurSeg,fp,0,0));
        end
      else se:=GetLastErrMsg;
    except
      on E:EWriteError do begin
        se:=E.Message;
        ec:=errFileWrite;
        end;
      end;
    if ec<>errOK then begin
      // set error flag on write errors
      FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
      raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
      end;
    end;
  end;

function TZipWrite.RemLastDir : string;
begin
  Result:='';
  with FileList do if Count>0 then begin
    with (Objects[Count-1] as TPkHeader) do if IsDir then begin
      Result:=Strings[Count-1];
      FStream.Size:=Offset;
      Free;
      Delete(Count-1);
      end;
    end;
  end;

// Add stream to archive (< 4GB, no encryption, no splitting)
procedure TZipWrite.AddStream (ASource : TStream; const AName : string; ACompress : boolean = true);
var
  ZfName    : RawByteString;
  ut,ct,fp  : int64;
  rf,fc     : cardinal;
  LFHeader  : TPKLocalHeader;
  ec        : integer;
  timo      : boolean;
  se        : string;
begin
  ut:=ASource.Size;
  with LFHeader do begin
    Signatur:=PKLocalSignatur;
    Flag:=0; ZfName:=AName;
    ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
    Vers:=VersDef;
    if (ASource.Size>0) and ACompress then begin
      Method:=zipDeflate;
      Flag:=Flag or zipMaxDeflFlag;
      end
    else Method:=zipNoComp;
    FTimeStamp:=FileTimeToFileAge(DateTimeToFileTime(Now));
    CRC:=0;
    FNLength:=length(ZfName);
    CSize:=0;
    USize:=0;
    ExtraLength:=0;
    end;
  fp:=FStream.Position;
  try
    FStream.Write(LFHeader,sizeof(TPKLocalHeader));         // local header
    FStream.Write(ZfName[1],LFHeader.FNLength);             // filename
    ZipThread:=TZipWriteThread.Create(ASource,nil,FStream,ACompress,false,not FIsTemp,
                                      FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      ec:=ErrorCode and errAllCodes;
      se:=SysErrorMsg;
      ut:=UTotal;
      ct:=CTotal;
      fc:=FCrc;
      Free;
      end;
    ZipThread:=nil;
    if ec=errOK then begin
      FStream.Position:=fp+14;                // position of crc
      FStream.Write(fc,sizeof(cardinal));
      // write 4 byte values
      rf:=TInt64(ct).Lo;
      FStream.Write(rf,sizeof(cardinal));
      rf:=TInt64(ut).Lo;
      FStream.Write(rf,sizeof(cardinal));
      FStream.Seek (0,soEnd);
      FileList.AddObject(AName,TPkHeader.Create(ZfName,0,0,LfHeader.FTimeStamp,fc,
               faArchive,LfHeader.Flag,LfHeader.Method,
               PkExtraCrypt.Version,PkExtraCrypt.Mode,PkExtraCrypt.Method,FCurSeg,fp,ct,ut));
      end;
  except
    on E:EWriteError do begin
      se:=E.Message;
      ec:=errFileWrite;
      end;
    end;
  if ec<>errOK then begin
    FStream.Size:=fp;    // reset on error
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
        raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

// Add file to archive
procedure TZipWrite.AddFile (const SourceName : string; var ARatio : Integer; ACompress : boolean = true);
begin
  AddFile(SourceName,ExtractRelativePath(FBaseDir,SourceName),ARatio,ACompress);
  end;

procedure TZipWrite.AddFile (const SourceName : string; ACompress : boolean = true);
var
  ARatio : integer;
begin
  AddFile(SourceName,ExtractRelativePath(FBaseDir,SourceName),ARatio,ACompress);
  end;

procedure TZipWrite.AddFile (const SourceName : string; FileData : TWin32FindData; ACompress : boolean = true);
var
  ARatio : integer;
begin
  AddFile(SourceName,ExtractRelativePath(FBaseDir,SourceName),FileData,ARatio,ACompress);
  end;

procedure TZipWrite.AddFileFrom (const SourceName,FilePath : string; FileData : TWin32FindData; var ARatio : Integer; ACompress : boolean = true);
begin
  AddFile(SourceName,ExtractRelativePath(FBaseDir,FilePath),FileData,ARatio,ACompress);
  end;

procedure TZipWrite.AddFile (const SourceName,FileName : string; ACompress : boolean = true);
var
  ARatio : integer;
begin
  AddFile(SourceName,FileName,ARatio,ACompress);
  end;

procedure TZipWrite.AddFile (const SourceName,FileName : string; var ARatio : Integer; ACompress : boolean = true);
var
  FileData : TWin32FindData;
  se       : string;
  ec       : integer;
begin
  if FileExistsXL(SourceName,false) then begin
    if GetFileDataXL(SourceName,FileData) then
        AddFile(SourceName,FileName,FileData,ARatio,ACompress)
    else begin
      ec:=errFileRead;
      se:=GetLastErrMsg;
      end;
    end
  else ec:=errNotFound;
  if ec<>errOK then begin
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

procedure TZipWrite.AddFile (const SourceName,FileName : string; FileData : TWin32FindData; var ARatio : Integer; ACompress : boolean = true);
var
  ZfName    : RawByteString;
  tzn,se    : string;
  ec        : integer;
  utime     : cardinal;
  LFHeader  : TPKLocalHeader;
  lef       : TPKExtraLocalField;
  n,nt,
  ut,ct,fp  : int64;
  z64,zae2  : boolean;
  rf,fc,el  : cardinal;
  sTmp      : TStream;
  timo,tf   : boolean;
  sm        : word;

  function ProcessStream (sDest : TStream) : integer;
  begin
    ZipThread:=TZipWriteThread.Create(SourceName,FEncryption,sDest,ACompress,false,not FIsTemp,
                                      FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      TempDir:=fZipTmpDir;
      repeat
        Application.ProcessMessages;
        Sleep(10);
        if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
        until Done or timo;
      Result:=ErrorCode and errAllCodes;
      se:=SysErrorMsg;
      ut:=UTotal;
      ct:=CTotal;
      fc:=FCrc;
      Free;
      end;
    ZipThread:=nil;
    end;

  procedure UpdateLocalHeader;
  begin
    if FEncrypt and zae2 then FStream.Position:=fp+18    // position of compressed size
    else begin                        // Crc if no encryption or AE-1
      FStream.Position:=fp+14;                // position of crc
      FStream.Write(fc,sizeof(cardinal));
      end;
    if z64 then begin
      // write 8 byte values to extra field
      FStream.Position:=fp+sizeof(TPKLocalHeader)+LFHeader.FNLength+4;
      FStream.Write (ut,sizeof(int64));
      FStream.Write (ct,sizeof(int64));
      end
    else begin
      // write 4 byte values
      rf:=TInt64(ct).Lo;
      FStream.Write(rf,sizeof(cardinal));
      rf:=TInt64(ut).Lo;
      FStream.Write(rf,sizeof(cardinal));
      end;
    FStream.Seek(0,soEnd);
    end;

begin
  se:='';
  utime:=GetFileAge(FileData);
  ut:=GetFileSize(FileData);
  z64:=ut>Z32MaxSize;
  zae2:=ut<AE1Size;
  with LFHeader do begin
    Signatur:=PKLocalSignatur;
    if IsOEMStr(FileName) then begin
      Flag:=0; ZfName:=FileName;
      ZfName:=StrToOEM(ZfName); // ZIP needs OEM characters (like DOS)
      end
    else begin
      Flag:=zipUtf8Flag;
      ZfName:=UTF8Encode(FileName);
      end;
    if z64 then Vers:=VersZip64 else Vers:=VersDef;
    if FEncrypt then begin
      Flag:=Flag or zipEncryptFlag;
      Method:=zipEncrypted;
      ACompress:=true;         // compress always on encryption
      end
    else begin
      if (ut>0) and ACompress then begin
        Method:=zipDeflate;
        Flag:=Flag or zipMaxDeflFlag;
        end
      else Method:=zipNoComp;
      end;
    FTimeStamp:=utime;
    CRC:=0;
    FNLength:=length(ZfName);
    if z64 then begin
      with lef do begin
        Tag:=ExtraZip64;
        ESize:=16;
        Buffer[0]:=0;
        Buffer[1]:=0;
        end;
      CSize:=MaxCardinal;
      USize:=MaxCardinal;
      el:=20;
      ExtraLength:=20;
      end
    else begin
      CSize:=0;
      USize:=0;
      el:=0;
      ExtraLength:=0;
      end;
    if FEncrypt then begin
      ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
      with PkExtraCrypt do begin
        if zae2 then Version:=VendorVersAE2 else Version:=VendorVersAE1;
        if ut>0 then Method:=zipDeflate else Method:=zipNoComp;
        end;
      end;
    end;
  try
    // check for space in segment
    n:=sizeof(TPKLocalHeader)+LFHeader.FNLength+el;
    if FEncrypt then inc(n,sizeof(TPKExtraEncField));
    if CheckSegment(n) then ec:=NewSegment else ec:=errOK;
    fp:=FStream.Position;
    FStream.Write(LFHeader,sizeof(TPKLocalHeader));       // local header
    FStream.Write(ZfName[1],LFHeader.FNLength);           // filename
    if z64 then FStream.Write(lef,el);                    // extra Zip64 field
    if FEncrypt then FStream.Write(PkExtraCrypt,sizeof(TPKExtraEncField));  // extra enc field
    if ec=errOK then begin
      sm:=FCurSeg;
      if fMaxSize>0 then begin   // temp. copy
        tf:=ut>UMaxMemSize;
        if tf then begin
          tzn:=FilenameToXL(fZipTmpDir+tmpZipName);
          sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
          end
        else sTmp:=TMemoryStream.Create;
        ec:=ProcessStream(sTmp);
        if ec=errOK then UpdateLocalHeader;
        sTmp.Position:=0;
        if ec=errOK then begin
          n:=fMaxSize-FStream.Position;
          if n<sTmp.Size then begin // split
            DoProgress(acCopy,-n,false);
            ec:=CopyStream(sTmp,FStream,n,nt,se);
            if ec=errOK then begin
              ec:=NewSegment;
              n:=sTmp.Size-n;
              while (ec=errOK) and (n>fMaxSize) do begin
                DoProgress(acCopy,-fMaxSize,false);
                ec:=CopyStream(sTmp,FStream,fMaxSize,nt,se);
                if ec=errOK then ec:=NewSegment;
                dec(n,fMaxSize);
                end;
              end;
            if ec=errOK then begin
              DoProgress(acCopy,-n,false);
              ec:=CopyStream(sTmp,FStream,n,nt,se);
              end;
            end
          else begin
            DoProgress(acCopy,-sTmp.Size,false);
            ec:=CopyStream(sTmp,FStream,sTmp.Size,nt,se);
            end;
          end;
        sTmp.Free;
        if tf then DeleteFile(tzn);
        end
      else begin
        ec:=ProcessStream(FStream);
        if ec=errOK then UpdateLocalHeader;
        end;
      end
    else se:=GetLastErrMsg;
    if ec=errOK then begin
      if ACompress and (ut>0) then begin
        ARatio:=round(1000*ct/ut);
        if ARatio>=1000 then ARatio:=-1;
        end
      else ARatio:=-1;
      FileList.AddObject(FileName,TPkHeader.Create(ZfName,0,0,utime,fc,
               FileGetAttrXL(SourceName),LfHeader.Flag,LfHeader.Method,
               PkExtraCrypt.Version,PkExtraCrypt.Mode,PkExtraCrypt.Method,sm,fp,ct,ut));
      end;
  except
    on E:EWriteError do begin
      se:=E.Message;
      ec:=errFileWrite;
      end;
    end;
  if (ec<>errOK) and (ec<>errFileFull) then FStream.Size:=fp;    // reset on error
  if ec<>errOK then begin
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec,se);
    end;
  end;

{ TPkHeader = class(TObject)
    FName           : RawByteString;   // Original filename (OEM or UTF8)
	  ErrCode,
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethod,
    Tag,DirLength,
    Segment,EncMode,
    EncVers,EncMeth : word;
    Offset,
    CSize,USize     : int64;
    IsDir,IsAE2,
    Zip64           : boolean;
}
procedure TZipWrite.SkipLocalHeader(sSource : TStream; AOfs : int64);
var
  LFHeader  : TPKLocalHeader;
  Buffer    : array of byte;
begin
  with sSource do begin
    Position:=AOfs;
    // Header berlesen
    Read(LFHeader,sizeof(TPKLocalHeader));         // local header
    SetLength(Buffer,LFHeader.FNLength);
    Read(Buffer[0],LFHeader.FNLength);             // filename length
    if LFHeader.ExtraLength>0 then begin
      SetLength(Buffer,LFHeader.ExtraLength);
      Read(Buffer[0],LFHeader.ExtraLength);        // filename
      end;
    end;
  end;

function TZipWrite.ProcessVerify (sSource : TStream; PkHeader : TPkHeader; var ErrMsg : string) : integer;
var
  timo      : boolean;
begin
  with PkHeader do
    ZipThread:=TZipVerifyThread.Create(FEncryption,Crc,sSource,CSize,CompMethod<>zipNoComp,false,not FIsTemp,
                      FBufferSize,FPriority,FOnProgress);
  with ZipThread do begin
    repeat
      Application.ProcessMessages;
      Sleep(10);
      if assigned(FOnTimeout) then timo:=FOnTimeout else timo:=false;
      until Done or timo;
    Result:=ErrorCode;
    ErrMsg:=SysErrorMsg;
    Free;
    end;
  end;

function TZipWrite.VerifyFile (PkHeader : TPkHeader; var ErrMsg : string) : integer;
begin
  with PkHeader do if CSize>0 then begin
    SkipLocalHeader(FStream,Offset);
    Result:=ProcessVerify(FStream,PkHeader,ErrMsg);
    end
  else Result:=errOK;
  end;

function TZipWrite.VerifySplittedFile(const ZName : string;
                        PkHeader : TPkHeader; var ErrMsg : string) : integer;
var
  tzn,szn  : string;
  sTmp  : TStream;
  sSeg  : TFileStream;
  tf    : boolean;
  seg   : word;
  n,nt  : int64;
begin
  if PkHeader.Segment=fCurSeg then Result:=VerifyFile(PkHeader,ErrMsg)
  else if PkHeader.CSize>0then  begin
    seg:=PkHeader.Segment;
    szn:=GetSegmentName(ZName,seg); //FilenameToXL(NewExt(ZName,'z'+Format('%.2d',[seg+1])));
    sSeg:=TFileStream.Create(szn,fmOpenRead);
    SkipLocalHeader(sSeg,PkHeader.Offset);
    if PkHeader.CSize+sSeg.Position<sSeg.Size then Result:=ProcessVerify(sSeg,PkHeader,ErrMsg)
    else begin        // need more than one segment
      tf:=PkHeader.USize>UMaxMemSize;
      if tf then begin
        tzn:=FilenameToXL(fZipTmpDir+tmpZipName);
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      DoProgress(acCopy,-PkHeader.CSize,false);
      n:=sSeg.Size-sSeg.Position;
      Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
      while (Result=errOK) and (sTmp.Size<PkHeader.CSize) do begin  // copy to sTmp
        Result:=OpenNextSegment(seg,sSeg);
        n:=PkHeader.CSize-n;
        if n>sSeg.Size then n:=sSeg.Size;
        if Result=errOK then Result:=CopyStream(sSeg,sTmp,n,nt,ErrMsg);
        end;
      sTmp.Position:=0;
      if Result=errOK then Result:=ProcessVerify(sTmp,PkHeader,ErrMsg);
      sTmp.Free;
      if tf then DeleteFile(tzn);
      end;
    if seg<>fCurSeg then FreeAndNil(sSeg);
    end
  else Result:=errOK;
  end;

procedure TZipWrite.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

procedure TZipWrite.Suspend;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Suspend;
  end;

procedure TZipWrite.Resume;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Resume;
  end;

{ TZipWriteThread ------------------------------------------------------------------- }
constructor TZipWriteThread.Create (const AFileName : string; AEncryption : TEncryption;
                               ZStream : TStream; ACompress,ASuspend,ASpeedLimit : boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TFileProgressEvent);
begin
  inherited Create (AFileName,'',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

constructor TZipWriteThread.Create (ASource : TStream; AEncryption : TEncryption;
                               ZStream : TStream; ACompress,ASuspend,ASpeedLimit : Boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TFileProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sSource:=ASource;
  sDest:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

{ ------------------------------------------------------------------- }
// add file to zip file
function TZipWriteThread.AddToZip : integer;
var
  ec : integer;
begin
  ec:=errOK;
  if length(SourceName)=0 then begin
    if not assigned(sSource) then ec:=errError+errStream;
    end
  else
    try
      sSource:=TFileStream.Create(SourceNameXL,fmOpenRead+fmShareDenyNone);
    except
      on E:EFOpenError do begin
        FSysErrMsg:=E.Message;
        ec:=errError+errFileOpen;
        end;
      end;
  if ec=errOK then begin
    ec:=ProcessStream(sSource,FCompress);
    try FreeAndNil(sSource); except end;
    end
  else if assigned(sSource) then try FreeAndNil(sSource); except end;
  Result:=ec;
  end;

procedure TZipWriteThread.Execute;
begin
  EndThread(AddToZip);
  end;

{ ------------------------------------------------------------------- }
constructor TPkFileRef.Create (Ofs : int64);
begin
  inherited Create;
  Offset:=Ofs;
  end;

{ TZipVerifyThread ------------------------------------------------------------------- }
constructor TZipVerifyThread.Create (AEncryption : TEncryption; ACrc : cardinal;
                   ZStream : TStream; ALength : int64; ACompress,ASuspend,ASpeedLimit : boolean;
                   ABufSize : integer; APriority : TThreadPriority; AOnProgress : TFileProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FSpeedLimit:=ASpeedLimit;
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FLength:=ALength;
  FCrc:=ACrc;
  FErrFlag:=errZip;
  end;

function TZipVerifyThread.VerifyZippedFile : integer;
var
  ec  : integer;
  crc : cardinal;
begin
  Crc:=FCrc;
  FCrc:=Crc32(0,nil,0);
  if assigned(FCrypting) then begin // decrypt file
    if not FCrypting.DecryptStream(sDest,nil,FLength,FSpeedLimit) then ec:=errError+errFileVerify
    else ec:=errOK;
    end
  else begin
    if FCompress then begin
      ec:=GzUncompress(sDest,nil,FLength,FSpeedLimit);
      end
    else ec:=CopyStream(sDest,nil,FLength,FSpeedLimit);
    if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
    end;
  Result:=ec;
  end;

procedure TZipVerifyThread.Execute;
begin
  EndThread(VerifyZippedFile);
  end;

{ TZipReadThread ------------------------------------------------------------------- }
constructor TZipReadThread.Create (ZStream : TStream; const AFileName : string;
                               AEncryption : TEncryption; AZipInfo : TZipFileInfo;
                               ASuspend : Boolean; ASpeedLimit : boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TFileProgressEvent);
begin
  inherited Create ('',AFileName,AEncryption,ASuspend,ABufSize,APriority);
  FSpeedLimit:=ASpeedLimit;
  FZipInfo:=AZipInfo;
  RdLocalHeader:=false;
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

constructor TZipReadThread.Create (ZStream,ADest : TStream;
                    AEncryption : TEncryption; AZipInfo : TZipFileInfo;
                    ASuspend,ASpeedLimit : boolean; ABufSize : integer;
                    APriority : TThreadPriority; AOnProgress : TFileProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  sDest:=ADest;
  FSpeedLimit:=ASpeedLimit;
  FZipInfo:=AZipInfo;
  RdLocalHeader:=false;
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

constructor TZipReadThread.Create (ZStream,ADest : TStream;
                    ASuspend,ASpeedLimit : boolean; ABufSize : integer;
                    APriority : TThreadPriority; AOnProgress : TFileProgressEvent);
begin
  inherited Create ('','',nil,ASuspend,ABufSize,APriority);
  sDest:=ADest;
  FSpeedLimit:=ASpeedLimit;
  RdLocalHeader:=true;
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

function TZipReadThread.CopyToStream (const AZipInfo : TZipFileInfo; fDest : TStream) : integer;
var
  sTmp          : TStream;
  tf            : boolean;
  tzn           : string;
begin
  if (AZipInfo.Flag and zipEncryptFlag)<>0 then begin  // decrypt
    if not assigned(FCrypting) then Result:=errError+errFileDCrypt
    else begin
      DoProgress(acDecrypt,-AZipInfo.CSize,FSpeedLimit);
      tf:=AZipInfo.CSize>CMaxMemSize;
      if tf then begin
        tzn:=FilenameToXL(fTmpDir+tmpEncName);
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      if FCrypting.DecryptStream(sSource,sTmp,AZipInfo.CSize,FSpeedLimit) then begin
        DoProgress(acUnCompress,-sTmp.Size,FSpeedLimit);
        sTmp.Seek(0,soFromBeginning);
        Result:=GzUncompress(sTmp,fDest,sTmp.size,FSpeedLimit);
        end
      else Result:=errError+errFileDCrypt;
      sTmp.Free;
      if tf then DeleteFile(tzn);
      end;
    end
  else begin
    DoProgress(acUnCompress,-AZipInfo.CSize,FSpeedLimit);
    if AZipInfo.Method=zipDeflate then Result:=GzUncompress(sSource,fDest,AZipInfo.CSize,FSpeedLimit)
    else Result:=CopyStream(sSource,fDest,AZipInfo.USize,FSpeedLimit);
    if (Result=errOK) and (FCrc<>AZipInfo.Crc) then Result:=errError+errFileRead;
    end;
  end;

function TZipReadThread.CopyToFile : integer;
begin
  Result:=errOK;
  if (length(DestName)=0) then begin  // no encryption
    if not assigned(sDest) then Result:=errError+errStream
    else if RdLocalHeader and not GetZipLocalHeader(sSource,FZipInfo) then Result:=errError+errFileRead;
    end
  else begin
    if FileExists(DestNameXL,false) then begin
      if FileSetAttr(DestNameXL,faArchive,false)<>0 then Result:=errError+errFileAttr;
      end;
    if Result=errOK then begin
      try
        sDest:=TWriteFileStream.Create(DestNameXL);
      except
        on E:Exception do begin
          FSysErrMsg:=E.Message;
          Result:=errError+errFileCreate;
          end;
        end;
      end;
    end;
  if (Result=errOK) then Result:=CopyToStream(FZipInfo,sDest);
  if (length(DestName)>0) then try FreeAndNil(sDest); except end;
  end;

procedure TZipReadThread.Execute;
begin
  EndThread(CopyToFile);
  end;

function TZipCompareThread.CompareToStream (const AZipInfo : TZipFileInfo; fDest : TStream) : integer;
var
  sTmp          : TStream;
  tf            : boolean;
  tzn           : string;
begin
  tf:=AZipInfo.CSize>CMaxMemSize;
  if tf then begin
    tzn:=FilenameToXL(fTmpDir+tmpGzipName);
    sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
    end
  else sTmp:=TMemoryStream.Create;
  Result:=CopyToStream (AZipInfo,sTmp);
  if Result=errOK then Result:=CompareStreams(sDest,sTmp,FSpeedLimit);
  sTmp.Free;
  if tf then DeleteFile(tzn);
  end;

function TZipCompareThread.CompareToFile : integer;
begin
  Result:=errOK;
  if (length(DestName)=0) then begin  // no encryption
    if not assigned(sDest) then Result:=errError+errStream
    else if RdLocalHeader and not GetZipLocalHeader(sSource,FZipInfo) then Result:=errError+errFileRead;
    end
  else begin
    if FileExists(DestNameXL,false) then begin
      try       // reference file
        sDest:=TFileStream.Create(DestNameXL,fmOpenRead+fmShareDenyNone);
      except
        on E:EFOpenError do begin
          FSysErrMsg:=E.Message;
          Result:=errError+errFileOpen;
          end;
        end;
      end
    else Result:=errError+errNotFound;
    end;
  if (Result=errOK) then Result:=CompareToStream(FZipInfo,sDest);
  try FreeAndNil(sDest); except end;
  end;

procedure TZipCompareThread.Execute;
begin
  EndThread(CompareToFile);
  end;

end.
